"""Sub-Module to contain all Formatted Postgres Commands & ORM Objs"""

from .orm import (
    Storable,
    AssetTable,
    TimeseriesConfig,
    MetadataArgs,
    METADATA_ARGS,
    MetadataInfo,
    DEFAULT_AGGREGATES,
)
from .enum import Operation, Schema, AssetTbls, SeriesTbls, GenericTbls
from .commands_obj import Commands, OperationMap
from .security import (
    SymbolArgs,
    SYMBOL_ARGS,
    STRICT_SYMBOL_ARGS,
)
from .timeseries import (
    TickArgs,
    AggregateArgs,
)


__all__ = (
    "Commands",
    "OperationMap",
    ###
    "Schema",
    "Operation",
    "AssetTbls",
    "SeriesTbls",
    "AssetTable",
    "GenericTbls",
    ###
    "Storable",
    "AssetTable",
    "MetadataInfo",
    "TimeseriesConfig",
    "DEFAULT_AGGREGATES",
    ####
    "MetadataArgs",
    "METADATA_ARGS",
    "TickArgs",
    "AggregateArgs",
    "SymbolArgs",
    "SYMBOL_ARGS",
    "STRICT_SYMBOL_ARGS",
)



from pandas import Timedelta, Timestamp
import pytest
from psyscale.dev import TimeseriesConfig, DEFAULT_AGGREGATES
from psyscale.psql.orm import AssetTable, _determine_conflicting_timedeltas

# pylint: disable=missing-function-docstring
STD_ASSET_LIST = ["us_fund", "us_stock", "crypto"]
CRYPTO_AGGS = [
    Timedelta("1h"),
    Timedelta("4h"),
    Timedelta("1D"),
    Timedelta("1W"),
]


TEST_CONFIG = TimeseriesConfig(
    STD_ASSET_LIST,  # type:ignore
    rth_origins={
        "us_stock": Timestamp("2000/01/03 08:30", tz="America/New_York"),
        "us_fund": Timestamp("2000/01/03 08:30", tz="America/New_York"),
    },
    eth_origins={
        "us_stock": Timestamp("2000/01/03 04:00", tz="America/New_York"),
        "us_fund": Timestamp("2000/01/03 04:00", tz="America/New_York"),
    },
    prioritize_rth={"us_stock": True, "us_fund": None},
    calculated_periods={"default": DEFAULT_AGGREGATES, "crypto": CRYPTO_AGGS},
    stored_periods={
        "default": [Timedelta("15s"), Timedelta("30s"), Timedelta("1m")],
        "crypto": [Timedelta("1m")],
    },
)


# region ---- ---- __post_init__ formatting  ---- ----


def test_for_errors():
    # Error on overlapping insertions & aggregates
    with pytest.raises(AttributeError):
        TimeseriesConfig(
            asset_classes=["stocks"],
            calculated_periods={"stocks": [Timedelta(days=1)]},
            stored_periods={"stocks": [Timedelta(days=1)]},
            rth_origins={},
            eth_origins={},
            htf_origins={},
        )

    with pytest.raises(KeyError):
        TEST_CONFIG.all_tables("bonds")


def test_config_asset_class_rth_true():
    assert TEST_CONFIG._ext_important("us_stock")
    # STD Tables
    tbls = TEST_CONFIG.std_tables("us_stock")
    assert set(tbl.asset_class for tbl in tbls) == {"us_stock"}
    assert sorted(tbl.period for tbl in tbls) == [
        Timedelta("5min"),
        Timedelta("15min"),
        Timedelta("30min"),
    ]

    tbls = TEST_CONFIG.std_tables("us_stock", inserted=True)
    assert set(tbl.asset_class for tbl in tbls) == {"us_stock"}
    assert sorted(tbl.period for tbl in tbls) == [
        Timedelta("15s"),
        Timedelta("30s"),
        Timedelta("1m"),
        Timedelta("5min"),
        Timedelta("15min"),
        Timedelta("30min"),
    ]

    # RTH Tables
    tbls = TEST_CONFIG.rth_tables("us_stock")
    assert set(tbl.asset_class for tbl in tbls) == {"us_stock"}
    assert sorted(tbl.period for tbl in tbls) == [
        Timedelta("1h"),
        Timedelta("4h"),
        Timedelta("1D"),
        Timedelta("1W"),
    ]

    tbls = TEST_CONFIG.rth_tables("us_stock", inserted=True)
    assert set(tbl.asset_class for tbl in tbls) == {"us_stock"}
    assert sorted(tbl.period for tbl in tbls) == [
        Timedelta("1h"),
        Timedelta("4h"),
        Timedelta("1D"),
        Timedelta("1W"),
    ]

    # ETH Tables
    tbls = TEST_CONFIG.eth_tables("us_stock")
    assert not set(tbl.asset_class for tbl in tbls)
    assert sorted(tbl.period for tbl in tbls) == []

    tbls = TEST_CONFIG.eth_tables("us_stock", inserted=True)
    assert not set(tbl.asset_class for tbl in tbls)
    assert sorted(tbl.period for tbl in tbls) == []

    # RAW Tables indirectly tested above


def test_config_asset_class_rth_none():
    assert TEST_CONFIG._ext_important("us_fund")
    # STD Tables
    tbls = TEST_CONFIG.std_tables("us_fund")
    assert set(tbl.asset_class for tbl in tbls) == {"us_fund"}
    assert sorted(tbl.period for tbl in tbls) == [
        Timedelta("5min"),
        Timedelta("15min"),
        Timedelta("30min"),
    ]

    tbls = TEST_CONFIG.std_tables("us_fund", inserted=True)
    assert set(tbl.asset_class for tbl in tbls) == {"us_fund"}
    assert sorted(tbl.period for tbl in tbls) == [
        Timedelta("15s"),
        Timedelta("30s"),
        Timedelta("1m"),
        Timedelta("5min"),
        Timedelta("15min"),
        Timedelta("30min"),
    ]

    # RTH Tables
    tbls = TEST_CONFIG.rth_tables("us_fund")
    assert set(tbl.asset_class for tbl in tbls) == {"us_fund"}
    assert sorted(tbl.period for tbl in tbls) == [
        Timedelta("1h"),
        Timedelta("4h"),
        Timedelta("1D"),
        Timedelta("1W"),
    ]

    tbls = TEST_CONFIG.rth_tables("us_fund", inserted=True)
    assert set(tbl.asset_class for tbl in tbls) == {"us_fund"}
    assert sorted(tbl.period for tbl in tbls) == [
        Timedelta("1h"),
        Timedelta("4h"),
        Timedelta("1D"),
        Timedelta("1W"),
    ]

    # ETH Tables
    tbls = TEST_CONFIG.eth_tables("us_fund")
    assert set(tbl.asset_class for tbl in tbls) == {"us_fund"}
    assert sorted(tbl.period for tbl in tbls) == [
        Timedelta("1h"),
        Timedelta("4h"),
        Timedelta("1D"),
        Timedelta("1W"),
    ]

    tbls = TEST_CONFIG.eth_tables("us_fund", inserted=True)
    assert set(tbl.asset_class for tbl in tbls) == {"us_fund"}
    assert sorted(tbl.period for tbl in tbls) == [
        Timedelta("1h"),
        Timedelta("4h"),
        Timedelta("1D"),
        Timedelta("1W"),
    ]

    # RAW Tables indirectly tested above


def test_config_asset_class_rth_indifferent():
    assert not TEST_CONFIG._ext_important("crypto")
    # STD Tables
    tbls = TEST_CONFIG.std_tables("crypto")
    assert set(tbl.asset_class for tbl in tbls) == {"crypto"}
    assert sorted(tbl.period for tbl in tbls) == [
        Timedelta("1h"),
        Timedelta("4h"),
        Timedelta("1D"),
        Timedelta("1W"),
    ]

    tbls = TEST_CONFIG.std_tables("crypto", inserted=True)
    assert set(tbl.asset_class for tbl in tbls) == {"crypto"}
    assert sorted(tbl.period for tbl in tbls) == [
        Timedelta("1m"),
        Timedelta("1h"),
        Timedelta("4h"),
        Timedelta("1D"),
        Timedelta("1W"),
    ]

    # RTH Tables
    tbls = TEST_CONFIG.rth_tables("crypto")
    assert not set(tbl.asset_class for tbl in tbls)
    assert sorted(tbl.period for tbl in tbls) == []

    tbls = TEST_CONFIG.rth_tables("crypto", inserted=True)
    assert not set(tbl.asset_class for tbl in tbls)
    assert sorted(tbl.period for tbl in tbls) == []

    # ETH Tables
    tbls = TEST_CONFIG.eth_tables("crypto")
    assert not set(tbl.asset_class for tbl in tbls)
    assert sorted(tbl.period for tbl in tbls) == []

    tbls = TEST_CONFIG.eth_tables("crypto", inserted=True)
    assert not set(tbl.asset_class for tbl in tbls)
    assert sorted(tbl.period for tbl in tbls) == []

    # RAW Tables indirectly tested above


# endregion

# region ---- ---- Method Testing  ---- ----


def test_get_selection_source_table():
    "Tests both get_selection_source_table and get_aggregation_source() that it relies on"
    # Exact Table is stored
    mock_table = AssetTable(asset_class="us_stock", period=Timedelta(days=1), rth=True, raw=False, ext=True)
    selected_table, needs_aggregation = TEST_CONFIG.get_selection_source_table(
        mock_table  # this func inherently allows for 'rtn_self'
    )
    assert needs_aggregation is False
    assert selected_table == mock_table

    # Test that the given table is not returned when flagged not too
    selected_table = TEST_CONFIG.get_aggregation_source(mock_table)
    assert selected_table != mock_table

    # Table needs to be aggregated from like RTH aggregated data
    mock_table = AssetTable(asset_class="us_stock", period=Timedelta(days=2), rth=True, raw=False, ext=True)
    selected_table, needs_aggregation = TEST_CONFIG.get_selection_source_table(mock_table)
    assert needs_aggregation
    assert selected_table == AssetTable(asset_class="us_stock", period=Timedelta(days=1), rth=True, raw=False, ext=True)

    # Table needs to be aggregated from a much lower timeframe that has needed EXT data
    mock_table = AssetTable(asset_class="us_stock", period=Timedelta(days=2), rth=False, raw=False, ext=True)
    selected_table, needs_aggregation = TEST_CONFIG.get_selection_source_table(mock_table)
    assert needs_aggregation
    assert selected_table == AssetTable(
        asset_class="us_stock", period=Timedelta("30min"), rth=None, raw=False, ext=True
    )

    # Test with invalid selections

    # No stored timeframe divisor
    with pytest.raises(AttributeError):
        mock_table = AssetTable(
            asset_class="us_stock",
            period=Timedelta("27s"),
            rth=True,
            raw=False,
            ext=True,
        )
        TEST_CONFIG.get_selection_source_table(mock_table)

    # Cannot Derive timeframe (lower than lowest stored)
    with pytest.raises(AttributeError):
        mock_table = AssetTable(
            asset_class="us_stock",
            period=Timedelta("1s"),
            rth=True,
            raw=False,
            ext=True,
        )
        TEST_CONFIG.get_selection_source_table(mock_table)


def test_get_tables_to_refresh():
    mock_table = AssetTable(asset_class="us_stock", period=Timedelta("1min"), rth=True, raw=True, ext=True)
    # Test All tables are returned
    assert TEST_CONFIG.get_tables_to_refresh(mock_table) == TEST_CONFIG.all_tables(
        mock_table.asset_class, include_raw=False
    )

    # test that only the tbls above the inserted timeframe are refreshed
    mock_table = AssetTable(asset_class="crypto", period=Timedelta("1h"), rth=True, raw=False, ext=True)
    tbls = TEST_CONFIG.all_tables(mock_table.asset_class, include_raw=False)
    tbls = [tbl for tbl in tbls if tbl.period > Timedelta("1h")]
    assert TEST_CONFIG.get_tables_to_refresh(mock_table) == tbls

    _ = """ideally the following test would also return true, but right now the function
    inst sophisticated enough to tell when a lower timeframe table doesn't actually feed
    an aggregated table. The edge case is too rare to care, and everything still functions
    as expected without the specificity."""
    # mock_table = AssetTable(
    #     asset_class="crypto", period=Timedelta("15s"), rth=True, raw=False, ext=True
    # )
    # # returns an empty list because all aggregates are formed on top of the
    # # raw '1min' timeframe table, not the '15s' table
    # assert TEST_CONFIG.get_tables_to_refresh(mock_table) == []


def test_config_from_table_names():

    # Shuffled Table names from the TEST_CONFIG definition from above
    test_config_table_names = [
        "_origin",  # should get filtered out
        "us_fund_15_raw_ext",
        "us_fund_30_raw_ext",
        "us_stock_30_raw_ext",
        "crypto_60_raw",
        "us_fund_3600_eth",
        "us_fund_14400_eth",
        "us_fund_86400_eth",
        "us_fund_604800_eth",
        "us_stock_60_raw_ext",
        "us_fund_300_ext",
        "us_fund_900_ext",
        "us_fund_1800_ext",
        "us_fund_3600_rth",
        "us_fund_60_raw_ext",
        "us_stock_15_raw_ext",
        "us_fund_14400_rth",
        "us_fund_86400_rth",
        "us_fund_604800_rth",
        "us_stock_300_ext",
        "us_stock_900_ext",
        "us_stock_1800_ext",
        "crypto_14400",
        "crypto_86400",
        "us_stock_3600_rth",
        "us_stock_14400_rth",
        "us_stock_86400_rth",
        "us_stock_604800_rth",
        "crypto_3600",
        "crypto_604800",
    ]

    reconstruct = TimeseriesConfig.from_table_names(test_config_table_names)

    for asset_class in reconstruct.asset_classes:
        assert reconstruct.all_tables(asset_class) == TEST_CONFIG.all_tables(asset_class)


# endregion


# region ---- ---- Conflicting time-delta tests  ---- ----


@pytest.mark.parametrize("prth", [True, False, None])
def test_determine_conflicting_timedeltas_no_conflict(prth):
    # Test case where RTH is prioritized
    rth_origin = Timestamp("2025-04-22 07:15:00")
    eth_origin = Timestamp("2025-04-22 07:15:00")
    periods = [Timedelta("15min"), Timedelta("30min"), Timedelta("1h")]

    std_periods, rth_periods, eth_periods = _determine_conflicting_timedeltas(
        periods, rth_origin, eth_origin, prioritize_rth=prth
    )

    # Returned Lists are always sorted smallest to largest
    assert std_periods == [Timedelta("15min"), Timedelta("30min"), Timedelta("1h")]
    assert rth_periods == []
    assert eth_periods == []


def test_determine_conflicting_timedeltas_with_conflict():
    rth_origin = Timestamp("2025-04-22 10:00:00")
    eth_origin = Timestamp("2025-04-22 07:15:00")
    periods = [Timedelta("15min"), Timedelta("30min"), Timedelta("1h")]

    # Test case where RTH is prioritized
    std_periods, rth_periods, eth_periods = _determine_conflicting_timedeltas(
        periods, rth_origin, eth_origin, prioritize_rth=True
    )

    assert std_periods == [Timedelta("15min")]
    assert rth_periods == [Timedelta("30min"), Timedelta("1h")]
    assert eth_periods == []

    # Test case where Neither are prioritized
    std_periods, rth_periods, eth_periods = _determine_conflicting_timedeltas(
        periods, rth_origin, eth_origin, prioritize_rth=False
    )

    assert std_periods == [Timedelta("15min")]
    assert rth_periods == []
    assert eth_periods == [Timedelta("30min"), Timedelta("1h")]

    # Test case where Eth is prioritized
    std_periods, rth_periods, eth_periods = _determine_conflicting_timedeltas(
        periods, rth_origin, eth_origin, prioritize_rth=None
    )

    assert std_periods == [Timedelta("15min")]
    assert rth_periods == [Timedelta("30min"), Timedelta("1h")]
    assert eth_periods == [Timedelta("30min"), Timedelta("1h")]


@pytest.mark.parametrize("prth", [True, False, None])
def test_invalid_eth_origin(prth):
    # Test invalid case: ETH origin occurs after RTH origin
    rth_origin = Timestamp("2025-04-22 08:00:00")
    eth_origin = Timestamp("2025-04-22 10:00:00")
    periods = [Timedelta("1h"), Timedelta("30min")]

    with pytest.raises(
        ValueError,
        match="TimescaleDB ETH Aggregate Origin must occur before RTH Origin.",
    ):
        _determine_conflicting_timedeltas(periods, rth_origin, eth_origin, prioritize_rth=prth)

    # Test invalid case: ETH and RTH origin difference >= 1 day
    rth_origin = Timestamp("2025-04-22 10:00:00")
    eth_origin = Timestamp("2025-04-21 10:00:00")

    with pytest.raises(
        ValueError,
        match="TimescaleDB RTH Origin Must be less than 1D after ETH Origin.",
    ):
        _determine_conflicting_timedeltas(periods, rth_origin, eth_origin, prioritize_rth=prth)


# endregion

MINUTE_CONFIG = TimeseriesConfig(
    ["equity"],  # type:ignore
    rth_origins={
        "equity": Timestamp("2000/01/03 08:30", tz="America/New_York"),
    },
    eth_origins={
        "equity": Timestamp("2000/01/03 04:00", tz="America/New_York"),
    },
    prioritize_rth={"equity": True},
    calculated_periods={"default": [Timedelta("5m"), Timedelta("30m"), Timedelta("1h")]},
    stored_periods={"default": [Timedelta("1m")]},
)


def test_no_origin_timestamp_conflict():
    hour_tbl = AssetTable("equity", Timedelta("1h"), False, True, True)
    rtn_tbl = MINUTE_CONFIG.get_aggregation_source(hour_tbl, rtn_self=False)

    # This should return the raw insterted minute table. This is because aggregates cannot be made
    # from continuous aggregates with different origin timestamps
    assert rtn_tbl == AssetTable("equity", Timedelta("1min"), True, True, None)



"Development package imports to allow for a more customizable Database implementation"

# pylint: disable=unused-wildcard-import, wildcard-import, unused-import
from psycopg import sql
from .psql.enum import *
from .psql.enum import Operation as Op
from .psql.generic import *
from .psql.orm import (
    Storable,
    AssetTable,
    MetadataInfo,
    TimeseriesConfig,
    DEFAULT_AGGREGATES,
    DEFAULT_ORIGIN_DATE,
    DEFAULT_HTF_ORIGIN_DATE,
)
from .psql.security import SymbolArgs, SYMBOL_ARGS, STRICT_SYMBOL_ARGS
from .psql.timeseries import TickArgs, TICK_ARGS, AggregateArgs, AGGREGATE_ARGS



"Symbols Table Partial Class Functions"

import logging
from typing import (
    Iterable,
    Literal,
    Any,
    Optional,
    Tuple,
)
from pandas import DataFrame, Series
from psycopg import sql
from psyscale.async_core import PsyscaleAsyncCore
from psyscale.core import PsyscaleConnectParams, PsyscaleCore

from .psql import (
    SYMBOL_ARGS,
    STRICT_SYMBOL_ARGS,
    SymbolArgs,
    Operation as Op,
    AssetTbls,
    Schema,
    GenericTbls,
)

# pylint: disable='protected-access'
log = logging.getLogger("psyscale_log")


class SymbolsPartial(PsyscaleCore):
    "Psyscale Symbols Table upsert and search functions"

    def __init__(
        self,
        conn_params: Optional[PsyscaleConnectParams | str] = None,
        *,
        down_on_del: bool = False,
        docker_compose_fpath: Optional[str] = None,
    ):
        # Chain the __init__ Docstring up the MRO since nothing changed
        self.__class__.__init__.__doc__ = super().__init__.__doc__
        super().__init__(
            conn_params,
            down_on_del=down_on_del,
            docker_compose_fpath=docker_compose_fpath,
        )
        self._ensure_securities_schema_format()

    def _ensure_securities_schema_format(self):
        with self._cursor() as cursor:
            cursor.execute(self[Op.SELECT, GenericTbls.SCHEMA_TABLES](Schema.SECURITY))
            tables: set[str] = {rsp[0] for rsp in cursor.fetchall()}

            if AssetTbls.SYMBOLS not in tables:
                # Init Symbols Table & pg_trgm Text Search Functions
                log.info("Creating Table '%s'.'%s'", Schema.SECURITY, AssetTbls.SYMBOLS)
                cursor.execute(self[Op.CREATE, AssetTbls._SYMBOL_SEARCH_FUNCS]())
                cursor.execute(self[Op.CREATE, AssetTbls.SYMBOLS]())

            # Init Symbol Data Range Metadata table & support function
            log.debug(
                "Ensuring Table '%s'.'%s' Exists",
                Schema.SECURITY,
                AssetTbls._METADATA,
            )
            cursor.execute(self[Op.CREATE, AssetTbls._METADATA_FUNC]())
            cursor.execute(self[Op.CREATE, AssetTbls._METADATA]())

    def upsert_securities(
        self,
        symbols: DataFrame,
        source: str,
        *,
        on_conflict: Literal["update", "ignore"] = "update",
    ) -> Tuple[Series, Series]:
        """
        Insert the Dataframe of symbols into the database.
        Primary Each for each entry is (Ticker, Exchange, Source)

        -- PARAMS --

        symbols: Dataframe.
            - Required Columns {ticker:str, name:str, exchange:str}:
                -- symbol:str - Ticker Symbol abbreviation
                -- name:Str - Full String Name of the Symbol
                -- exchange:str - Abbrv. Exchange Name for the symbol
                -- asset_class:str - Type of asset, This must match the Timeseries Config
                    asset_classes otherwise data for this symbol cannot be inserted into the
                    database.

            - Any Extra Columns will be packaged into JSON and dumped into an 'attrs' Dictionary.
            To prevent bloat, drop all columns that will not be used.

        source: string.
            - string representation of what API Sourced the symbol
            - e.g. Alpaca, IBKR, Polygon, Coinbase, etc.

        on_conflict: Literal["update", "ignore"] : default = "update"
            Update or Do Nothing when given a symbol that has a conflicting primary key.

        -- RETURNS --
            Tuple of [Series, Series]
            First Series Object is inserted Symbols
            Second Series Object is Updated Symbols
        """
        symbols_fmt = _prep_symbols_upsert(symbols, source)
        if symbols_fmt is None:
            return Series(), Series()

        with self._cursor() as cursor:
            # Create & Inject the Data into a Temporary Table
            cursor.execute(self[Op.CREATE, AssetTbls.SYMBOLS_BUFFER]())
            copy_cmd = self[Op.COPY, AssetTbls.SYMBOLS_BUFFER](
                # Sends the COPY Cmd & the order of the Columns of the Dataframe
                [str(c) for c in symbols_fmt.columns]
            )
            with cursor.copy(copy_cmd) as copy:
                for row in symbols_fmt.itertuples(index=False, name=None):
                    # Writes each row as a Tuple that matches the Dataframe Column Order
                    copy.write_row(row)

            # Merge the Temp Table By inserting / upserting from the Temporary Table
            _op = Op.UPSERT if on_conflict == "update" else Op.INSERT
            cursor.execute(self[_op, AssetTbls.SYMBOLS_BUFFER](source))
            response = DataFrame(cursor.fetchall())

        return _pkg_symbols_upsert_response(response, _op)

    def search_symbols(
        self,
        filter_args: dict[SymbolArgs | str, Any],
        return_attrs: bool = False,
        attrs_search: bool = False,
        limit: int | None = 100,
        *,
        strict_symbol_search: bool | Literal["ILIKE", "LIKE", "="] = False,
    ) -> list[dict]:
        """
        Search the database's symbols table returning all the symbols that match the given criteria.
        Search function supports trigram based fuzzy name + symbol search.

        -- PARAMS --
        - filter_args: dict[SymbolArgs | str : Any]
            - The filtering arguments that need to be matched against. By default only the keys that
            match the table column names (SymbolArgs Literal, e.g. pkey, name, symbol, etc.) will be
            used.
            - All Arguments aside from 'name' and 'symbol' will be used in a strict '=' comparison
            filter. 'name' will always be used in a fuzzystr trigram search where the results are
            ordered by relevancy. by default, 'symbol' will also be a fuzzystr trigram search.
            - When a 'pkey' filter is given, all other filter keys are ignored and the table is
            searched for the given integer pkey. This is because, by table definition, the pkey
            will be unique and can only ever return a single row.
            - Additional argument keys  that are not column names of the table (Not in SymbolArgs
            Literal) will be ignored by default. See attrs_search for more on this behavior.

        - return_attrs: boolean.
            - When True return an 'attrs' dictionary that has all additional attributes of the
            symbol that are stored in the 'attrs' column of the table.

        - attrs_search: boolean.
            - When True any additional keys that are given as filters args, but not recognized as
            table columns, will be used in a strict search against that 'attrs' JSON Column of the
            table.
            - When False additional keys within the filter_args are ignored.
            - i.e. when true, if {'shortable':True} is passed in filter_args then only rows that
            have a defined 'shortable'= True Attrubute will be returned.
            - This search will only ever be a strict '=' comparison, so if searching for a string
            or int the value given must be exact.

        - limit: int
            - The Optional Integer limit of symbol results to return.

        - strict_symbol_search: Boolean | Literal["ILIKE", "LIKE", "="] : default False.
            - When False a fuzzystr trigram search is used and the results are ordered by relevancy.
            Even if an exact match for the symbol is returned, this will still result in other
            similar symbols being returned.
            - When not False the symbol search will use the given PostgreSQL comparator. True
            equates to passing 'ILIKE' Which ignores case.
            - This is far more useful when passing a symbol with wildcard args. e.g.
            'ILIKE' + {symbol:'sp'} will likely not return results, 'ILIKE' + {symbol:'sp%'}
            will return all symbols starting with 'SP', case insensitive.
        """

        name, symbol, limit, filters, attrs = _prep_symbol_search(
            filter_args, attrs_search, limit, strict_symbol_search
        )

        with self._cursor(dict_cursor=True) as cursor:
            cursor.execute(self[Op.SELECT, AssetTbls.SYMBOLS](name, symbol, filters, return_attrs, attrs, limit))
            return cursor.fetchall()

    def distinct_sources(self, filters: dict[SymbolArgs | str, Any] = {}) -> list[str]:
        "Return the distinct Data Broker Sources stored in the Symbols Table."
        return [arg[0] for arg in self.select_distinct_symbols({"source"}, filters)]

    def distinct_exchanges(self, filters: dict[SymbolArgs | str, Any] = {}) -> list[str]:
        "Return the Distinct Exchanges stored in the Symbols Table."
        return [arg[0] for arg in self.select_distinct_symbols({"exchange"}, filters)]

    def distinct_asset_classes(self, filters: dict[SymbolArgs | str, Any] = {}) -> list[str]:
        "Return the distinct Asset Classes stored in the Symbols Table."
        return [arg[0] for arg in self.select_distinct_symbols({"asset_class"}, filters)]

    def select_distinct_symbols(
        self, args: Iterable[SymbolArgs], filters: dict[SymbolArgs | str, Any] = {}
    ) -> list[tuple]:
        "Select the distinct combinations of the given arguments from the Symbols Table."
        args = [arg for arg in args if arg in SYMBOL_ARGS]  # filter to only valid args
        _filters = [(k, "=", v) for k, v in filters.items()]
        rsp, _ = self.execute(
            self[Op.SELECT, GenericTbls.TABLE](Schema.SECURITY, AssetTbls.SYMBOLS, args, _filters, distinct=True),
        )
        return rsp

    def update_symbol(self, symbols: int | str | list[int | str], args: dict[SymbolArgs | str, Any]) -> bool:
        """
        Update a Single Symbol or list of symbols with the given arguments.

        This method is remained general for utility purposes. It should generally just be used to
        update the stored_[tick/minute/aggregate] columns. All Other parameters should remain
        constant by nature. To insert symbols see the API_Extension that allows this to be done
        in bulk.

        Note: Setting any stored column = False does not Delete any Data.

        :params:
        - symobls : int, str, or list[int | str]
            - Identifying symbol (str) or primary key (int) of the symbol to update.
            May be a single value or list of values.
            - Interger Pkeys are preferred method since multiple rows of the same symbol may be
            inserted into the database. Unique Symbol collisions only occur on conflicting
            (symbol, exchange, source) combinations while pkeys are always unique.

        - args : Dict[SymbolArgs, Any]
            - A Dictionary of Column Values to update. If PKEY is passed as a Key it will be
            ignored. Extra Keys are Ignored.
            - Note: This can throw a psycopg.Database Error if passed an update to Symbol, Source,
            or Exchange that would result in a change that would violate the UNIQUE flag on those
            collective parameters.

        :returns: Boolean, True on Successful Update.
        """
        # Convert varierty of symbol inputs to consistent list of integer pkeys
        if isinstance(symbols, int):
            pkeys = [symbols]
        elif isinstance(symbols, str):
            pkeys = [self._get_pkey(symbols)]
            if pkeys[0] is None:
                log.warning("Cannot Update Symbol %s, symbol is not known.", symbols)
                return False
        else:
            _tmp_list = [(self._get_pkey(symbol), symbol) for symbol in symbols]
            pkeys = [_tmp_list for pkey, _ in _tmp_list if pkey is not None]
            unknown_symbols = [symbol for pkey, symbol in _tmp_list if pkey is None]
            log.warning("Cannot Update Symbol(s) %s, symbol is not known.", unknown_symbols)

        if "pkey" in args:
            args.pop("pkey")

        _args = [(k, v) for k, v in args.items() if k in SYMBOL_ARGS]
        if len(_args) == 0:
            log.warning(
                "Attemping to update Database symbol but no arg updates where given. pkey(s) = %s",
                pkeys,
            )
            return False

        if len(pkeys) == 0:
            log.warning("Attemping to update Database symbol(s) but no pkeys were given")
            return False

        # Convert The pkeys to a List so only one Update Command needs to be sent.
        _filter = sql.SQL("pkey=ANY(ARRAY[{pkeys}])").format(pkeys=sql.SQL(",").join(sql.Literal(v) for v in pkeys))

        with self._cursor() as cursor:
            cursor.execute(self[Op.UPDATE, AssetTbls.SYMBOLS](_args, _filter))
            return cursor.statusmessage is not None and cursor.statusmessage == "UPDATE 1"

        return False  # Default return if cursor throws error


class AsyncSymbolsPartial(PsyscaleAsyncCore, SymbolsPartial):
    "Async Symbols Table search function"

    async def upsert_securities_async(
        self,
        symbols: DataFrame,
        source: str,
        *,
        on_conflict: Literal["update", "ignore"] = "update",
    ) -> Tuple[Series, Series]:
        """
        Insert the Dataframe of symbols into the database.
        Primary Each for each entry is (Ticker, Exchange, Source)

        -- PARAMS --

        symbols: Dataframe.
            - Required Columns {ticker:str, name:str, exchange:str}:
                -- symbol:str - Ticker Symbol abbreviation
                -- name:Str - Full String Name of the Symbol
                -- exchange:str - Abbrv. Exchange Name for the symbol
                -- asset_class:str - Type of asset, This must match the Timeseries Config
                    asset_classes otherwise data for this symbol cannot be inserted into the
                    database.

            - Any Extra Columns will be packaged into JSON and dumped into an 'attrs' Dictionary.
            To prevent bloat, drop all columns that will not be used.

        source: string.
            - string representation of what API Sourced the symbol
            - e.g. Alpaca, IBKR, Polygon, Coinbase, etc.

        on_conflict: Literal["update", "ignore"] : default = "update"
            Update or Do Nothing when given a symbol that has a conflicting primary key.

        -- RETURNS --
            Tuple of [Series, Series]
            First Series Object is inserted Symbols
            Second Series Object is Updated Symbols
        """
        symbols_fmt = _prep_symbols_upsert(symbols, source)
        if symbols_fmt is None:
            return Series(), Series()

        async with self._acursor() as cursor:
            # Create & Inject the Data into a Temporary Table
            await cursor.execute(self[Op.CREATE, AssetTbls.SYMBOLS_BUFFER]())
            copy_cmd = self[Op.COPY, AssetTbls.SYMBOLS_BUFFER](
                # Sends the COPY Cmd & the order of the Columns of the Dataframe
                [str(c) for c in symbols_fmt.columns]
            )
            async with cursor.copy(copy_cmd) as copy:
                for row in symbols_fmt.itertuples(index=False, name=None):
                    # Writes each row as a Tuple that matches the Dataframe Column Order
                    await copy.write_row(row)

            # Merge the Temp Table By inserting / upserting from the Temporary Table
            _op = Op.UPSERT if on_conflict == "update" else Op.INSERT
            await cursor.execute(self[_op, AssetTbls.SYMBOLS_BUFFER](source))
            response = DataFrame(await cursor.fetchall())

        return _pkg_symbols_upsert_response(response, _op)

    async def search_symbols_async(
        self,
        filter_args: dict[SymbolArgs | str, Any],
        return_attrs: bool = False,
        attrs_search: bool = False,
        limit: int | None = 100,
        *,
        strict_symbol_search: bool | Literal["ILIKE", "LIKE", "="] = False,
    ) -> list[dict]:
        "See symbol_search method docstring"
        name, symbol, limit, filters, attrs = _prep_symbol_search(
            filter_args, attrs_search, limit, strict_symbol_search
        )

        async with self._acursor(dict_cursor=True) as cursor:
            await cursor.execute(self[Op.SELECT, AssetTbls.SYMBOLS](name, symbol, filters, return_attrs, attrs, limit))
            return await cursor.fetchall()


def _prep_symbol_search(
    filter_args: dict[SymbolArgs | str, Any],
    attrs_search: bool,
    limit: int | None,
    strict_symbol_search: bool | Literal["ILIKE", "LIKE", "="],
) -> tuple:

    if "pkey" in filter_args:
        # Fast Track PKEY Searches since there will only ever be 1 symbol returned
        filters = [("pkey", "=", filter_args["pkey"])]
        name, symbol, attrs, limit = None, None, None, 1
    else:
        filters = [(k, "=", v) for k, v in filter_args.items() if k in STRICT_SYMBOL_ARGS]

        name = filter_args.get("name", None)
        symbol = filter_args.get("symbol", None)

        if strict_symbol_search and symbol is not None:
            # Determine if symbol is passed as a strict or fuzzy parameter
            compare_method = (
                strict_symbol_search
                if isinstance(strict_symbol_search, str)
                else "ILIKE"  # Default search for similar symbols that match ignoring case.
            )
            filters.append(("symbol", compare_method, symbol))
            symbol = None  # Prevents the 'similarity' search from being added

        # Filter all extra given filter params into a separate dict
        attrs = dict((k, v) for k, v in filter_args.items() if k not in SYMBOL_ARGS) if attrs_search else None
        if attrs and len(attrs) == 0:
            attrs = None

    return name, symbol, limit, filters, attrs


def _prep_symbols_upsert(symbols: DataFrame, source: str) -> DataFrame | None:

    if not isinstance(source, str) or source == "":
        log.error("Cannot Insert Symbols, Invalid Source, Source = %s", source)
        return
    if not isinstance(symbols, DataFrame):
        log.error("Cannot Insert Symbols, Invalid Symbols Argument")
        return

    symbols.columns = symbols.columns.str.lower()
    req_cols = {"symbol", "name", "exchange", "asset_class"}
    missing_cols = req_cols.difference(symbols.columns)
    if len(missing_cols) != 0:
        log.error("Cannot insert symbols. Dataframe missing Columns: %s", missing_cols)
        return

    # Convert to a format that can be inserted into the database
    symbols_fmt = symbols[[*req_cols]].copy()

    # Turn all extra Columns into an attributes json obj.
    symbols_fmt.loc[:, "attrs"] = symbols[[*set(symbols.columns).difference(req_cols)]].apply(
        lambda x: x.to_json(), axis="columns"
    )

    return symbols_fmt


def _pkg_symbols_upsert_response(response: DataFrame, _op: Op) -> Tuple[Series, Series]:
    if len(response) == 0:
        return Series(), Series()

    if _op == Op.INSERT:
        # All Returned symbols in response were inserted, none updated.
        return response[0], Series()
    else:
        # Second Column is xmax, on insertion this is 0, on update its != 0
        inserted = response[1] == "0"
        return response.loc[inserted, 0], response.loc[~inserted, 0]



import pytest
import pandas as pd
from psyscale.dev import sql, Schema, AssetTbls

# region ---- ---- Test Fixtures ---- ----


@pytest.fixture
def symbols_df():
    return pd.DataFrame(
        [
            {
                "symbol": "AAPL",
                "name": "Apple Inc.",
                "exchange": "NASDAQ",
                "asset_class": "equity",
                "sector": "Technology",
            },
            {
                "symbol": "GOOG",
                "name": "Alphabet Inc.",
                "exchange": "NASDAQ",
                "asset_class": "equity",
                "sector": "Technology",
            },
            {
                "symbol": "SPY",
                "name": "S&P 500",
                "exchange": "ARCA",
                "asset_class": "fund",
                "sector": "Something",
            },
        ]
    )


@pytest.fixture
def attrs_symbols():
    return pd.DataFrame(
        [
            {
                "symbol": "TEST1",
                "name": "Test Symbol One",
                "exchange": "NYSE",
                "asset_class": "equity",
                "sector": "Technology",  # Extra column → goes into `attrs`
                "shortable": True,
            },
            {
                "symbol": "TEST2",
                "name": "Test Symbol Two",
                "exchange": "NASDAQ",
                "asset_class": "equity",
                "sector": "Healthcare",
                "shortable": False,
            },
        ]
    )


def _clean(psyscale_db):
    psyscale_db.execute(
        sql.SQL("TRUNCATE TABLE {schema}.{table} RESTART IDENTITY CASCADE;").format(
            schema=sql.Identifier(Schema.SECURITY),
            table=sql.Identifier(AssetTbls.SYMBOLS),
        )
    )


@pytest.fixture
def clean_symbols_table(psyscale_db):
    """Truncates the symbols table before & after each test run"""
    _clean(psyscale_db)
    yield
    _clean(psyscale_db)


# endregion

# region ---- ---- Test Basic Symbol Insert & Search ---- ----


def test_upsert_inserts_symbols(psyscale_db, clean_symbols_table, symbols_df):
    inserted, updated = psyscale_db.upsert_securities(symbols_df, source="test_api")

    assert set(inserted.tolist()) == {"AAPL", "GOOG", "SPY"}
    assert updated.empty


def test_select_distinct_symbols(psyscale_db, clean_symbols_table, symbols_df):
    psyscale_db.upsert_securities(symbols_df, source="test_api")

    equal_sets = lambda x, y: set(x).issubset(y) and set(x).issuperset(y)

    assert equal_sets(psyscale_db.distinct_sources(), ["test_api"])
    assert equal_sets(psyscale_db.distinct_exchanges(), ["NASDAQ", "ARCA"])
    assert equal_sets(psyscale_db.distinct_asset_classes(), ["equity", "fund"])

    combos = psyscale_db.select_distinct_symbols(["asset_class", "exchange"])

    assert ("equity", "NASDAQ") in combos
    assert ("fund", "ARCA") in combos


def test_upsert_ignore_conflict(psyscale_db, clean_symbols_table, symbols_df):
    # Initial insert
    psyscale_db.upsert_securities(symbols_df, source="test_api")

    # Modify names to test if they're ignored
    modified = symbols_df.copy()
    modified.loc[modified.symbol == "AAPL", "name"] = "Fake Apple Inc."

    inserted, updated = psyscale_db.upsert_securities(modified, source="test_api", on_conflict="ignore")

    assert inserted.empty
    assert updated.empty

    # Fetch from db and confirm original value remains
    result = psyscale_db.search_symbols({"symbol": "AAPL", "source": "test_api"}, strict_symbol_search="=")
    assert result[0]["name"] == "Apple Inc."  # Original name retained

    psyscale_db.execute(
        sql.SQL("TRUNCATE TABLE {schema}.{table};").format(schema=Schema.SECURITY, table=AssetTbls.SYMBOLS)
    )


def test_upsert_updates_on_conflict(psyscale_db, clean_symbols_table, symbols_df):
    # Initial insert
    psyscale_db.upsert_securities(symbols_df, source="test_api")

    # Confirm the insert
    result = psyscale_db.search_symbols({"symbol": "GOOG", "source": "test_api"}, strict_symbol_search="=")
    assert result[0]["name"] == "Alphabet Inc."

    # Modify one symbol to simulate an update
    modified = symbols_df.copy()
    modified.loc[modified.symbol == "GOOG", "name"] = "Alphabet Inc. Class A"

    inserted, updated = psyscale_db.upsert_securities(modified, source="test_api", on_conflict="update")

    assert inserted.empty
    assert updated.tolist() == ["AAPL", "GOOG", "SPY"]

    # Confirm the update
    result = psyscale_db.search_symbols({"symbol": "GOOG", "source": "test_api"}, strict_symbol_search="=")
    assert result[0]["name"] == "Alphabet Inc. Class A"

    psyscale_db.execute(
        sql.SQL("TRUNCATE TABLE {schema}.{table};").format(schema=Schema.SECURITY, table=AssetTbls.SYMBOLS)
    )


# endregion

# region ---- ---- Test symbol Atters Insert & Search ---- ----


def test_upsert_with_attrs_and_retrieval(psyscale_db, clean_symbols_table, attrs_symbols):
    inserted, updated = psyscale_db.upsert_securities(attrs_symbols, source="UnitTest")

    # Validate both were inserted
    assert set(inserted.tolist()) == {"TEST1", "TEST2"}
    assert updated.empty

    # Validate full retrieval including attrs
    for symbol in ["TEST1", "TEST2"]:
        result = psyscale_db.search_symbols(
            {
                "symbol": symbol,
                "exchange": attrs_symbols.loc[attrs_symbols["symbol"] == symbol, "exchange"].iloc[0],
            },
            return_attrs=True,
            strict_symbol_search="=",
        )
        assert len(result) == 1
        row = result[0]

        assert row["symbol"] == symbol
        assert "attrs" in row
        assert row["attrs"].get("sector") == attrs_symbols.loc[attrs_symbols["symbol"] == symbol, "sector"].iloc[0]
        assert (
            row["attrs"].get("shortable") == attrs_symbols.loc[attrs_symbols["symbol"] == symbol, "shortable"].iloc[0]
        )


def test_search_by_jsonb_attr(psyscale_db, clean_symbols_table, attrs_symbols):
    # Ensure test data is inserted
    psyscale_db.upsert_securities(attrs_symbols, source="UnitTest")

    # Search for a symbol by JSONB attribute
    results = psyscale_db.search_symbols({"shortable": True}, return_attrs=True, attrs_search=True)

    assert isinstance(results, list)
    assert any(row["symbol"] == "TEST1" for row in results)
    assert all("attrs" in row for row in results)
    assert all(row["attrs"].get("shortable") is True for row in results)


# endregion


def test_fuzzy_trigram_search_orders_results(psyscale_db, clean_symbols_table):
    test_data = pd.DataFrame(
        [
            {
                "symbol": "TST",
                "name": "Test Symbol",
                "exchange": "NYSE",
                "asset_class": "equity",
            },
            {
                "symbol": "TST1",
                "name": "Test Symbol 1",
                "exchange": "NYSE",
                "asset_class": "equity",
            },
            {
                "symbol": "TST2",
                "name": "Testing Simbol",
                "exchange": "NYSE",
                "asset_class": "equity",
            },
            {
                "symbol": "TXS",
                "name": "Tex Symbol",
                "exchange": "NYSE",
                "asset_class": "equity",
            },
        ]
    )

    psyscale_db.upsert_securities(test_data, source="UnitTest")

    # Fuzzy symbol search
    results = psyscale_db.search_symbols({"symbol": "tst"}, strict_symbol_search=False)

    assert len(results) == 4  # All Results are returned
    symbols = [r["symbol"] for r in results]

    # Check that the most similar match ("TST") is first
    assert symbols[0] == "TST"
    # Check that the least similar match ("TXS") is last
    assert symbols[-1] == "TXS"

    # Fuzzy name search
    results = psyscale_db.search_symbols({"name": "symbol"}, strict_symbol_search=False)
    names = [r["name"] for r in results]

    assert len(names) == 4  # All Results are returned
    # Most relevant has desired result & the fewest extra characters
    assert names[0] == "Tex Symbol"
    assert names[1] == "Test Symbol"
    assert names[2] == "Test Symbol 1"
    assert names[3] == "Testing Simbol"  # Least Relevant Should be last.


def test_update_symbol_valid_and_constraint_violation(psyscale_db, symbols_df, clean_symbols_table, caplog):
    # Insert test symbols
    psyscale_db.upsert_securities(symbols_df, source="UnitTest")

    # Retrieve pkey of "AAPL"
    aapl = psyscale_db.search_symbols({"symbol": "AAPL", "source": "UnitTest"}, strict_symbol_search="=")
    assert aapl and "pkey" in aapl[0]
    pkey = aapl[0]["pkey"]

    # Valid update
    result = psyscale_db.update_symbol(pkey, {"store_tick": True})
    assert result is True

    # Ensure update can be retrieved (and pkey search works)
    aapl = psyscale_db.search_symbols({"pkey": pkey}, strict_symbol_search="=", return_attrs=True)[0]
    assert "store_tick" in aapl and aapl["store_tick"]
    assert "store" in aapl and aapl["store"]

    # Check Symbols table constraint shows error
    with caplog.at_level("ERROR"):
        result = psyscale_db.update_symbol(pkey, {"store_minute": True})
        assert result is False
    with caplog.at_level("ERROR"):
        result = psyscale_db.update_symbol(pkey, {"store_aggregate": True})
        assert result is False

    # Optional: Confirm original value was not modified post-exception
    updated = psyscale_db.search_symbols({"pkey": pkey})[0]
    assert updated["store_tick"] is True
    assert updated["store_minute"] is False
    assert updated["store_aggregate"] is False



from psycopg import DatabaseError
import pytest
import pandas as pd
from pandas.testing import assert_series_equal, assert_frame_equal
import numpy as np

from psyscale.dev import TimeseriesConfig
from psyscale.dev import Schema
from psyscale.psql.timeseries import AGGREGATE_ARGS
from psyscale.series_df import Series_DF
from psyscale import PsyscaleDB

# region ---- ---- Test Fixtures ---- ----


@pytest.fixture
def AAPL_MIN_DATA():
    df = pd.read_csv("example_data/AAPL_1min.csv")
    df.rename({"dt": "date", "open": "o", "high": "max"}, inplace=True)
    yield df


@pytest.fixture
def TICK_DATA():
    yield pd.read_csv("example_data/example_ticks.csv")


MINUTE_CONFIG = TimeseriesConfig(
    ["equity"],  # type:ignore
    rth_origins={
        "equity": pd.Timestamp("2000/01/03 08:30", tz="America/New_York"),
    },
    eth_origins={
        "equity": pd.Timestamp("2000/01/03 04:00", tz="America/New_York"),
    },
    prioritize_rth={"equity": True},
    calculated_periods={"default": [pd.Timedelta("5m"), pd.Timedelta("30m"), pd.Timedelta("1h")]},
    stored_periods={"default": [pd.Timedelta("1m")]},
)


TICK_CONFIG = TimeseriesConfig(
    ["equity"],  # type:ignore
    rth_origins={
        "equity": pd.Timestamp("2000/01/03 08:30", tz="America/New_York"),
    },
    eth_origins={
        "equity": pd.Timestamp("2000/01/03 04:00", tz="America/New_York"),
    },
    prioritize_rth={"equity": True},
    calculated_periods={"default": [pd.Timedelta("15s"), pd.Timedelta("30s")]},
    stored_periods={"default": [pd.Timedelta(0)]},
)


@pytest.fixture(scope="module")
def psyscale_db(test_url):
    "Module level PsyscaleMod that inits a couple symbols and the needed timeseries config"
    db = PsyscaleDB(test_url)
    db.configure_timeseries_schema(minute_tables=MINUTE_CONFIG, tick_tables=TICK_CONFIG)
    db.upsert_securities(
        pd.DataFrame(
            [
                {
                    "symbol": "AAPL",
                    "name": "Apple Inc.",
                    "exchange": "NASDAQ",
                    "asset_class": "equity",
                    "sector": "Technology",
                },
                {
                    "symbol": "GOOG",
                    "name": "Alphabet Inc.",
                    "exchange": "NASDAQ",
                    "asset_class": "equity",
                    "sector": "Technology",
                },
            ]
        ),
        source="unit_test",
    )
    aapl = db.search_symbols({"symbol": "AAPL"}, strict_symbol_search=True)[0]
    db.update_symbol(aapl["pkey"], {"store_minute": True})
    db.update_symbol("goog", {"store_tick": True})
    yield db


def test_01_metadata_fetch(psyscale_db: PsyscaleDB):
    symbols_to_insert = psyscale_db.search_symbols({"store": True})
    assert len(symbols_to_insert) == 2

    aapl_search = psyscale_db.search_symbols({"store_minute": True})
    assert len(aapl_search) == 1
    aapl = aapl_search[0]
    assert aapl["symbol"] == "AAPL"
    assert "pkey" in aapl
    aapl_pkey = aapl["pkey"]

    # No Data inserted yet, MetaData should be empty
    metadata = psyscale_db.stored_metadata(aapl_pkey)
    assert len(metadata) == 0

    # Should show that we need to store data in the Minute table of minute schema
    metadata = psyscale_db.stored_metadata(aapl_pkey, _all=True)
    assert len(metadata) == 1
    metadata = metadata[0]

    assert metadata.schema_name == Schema.MINUTE_DATA
    assert metadata.start_date == metadata.end_date
    assert metadata.start_date == pd.Timestamp("1800-01-01", tz="UTC")
    assert metadata.timeframe == pd.Timedelta("1min")


# region ---- ---- Aggregate Data Tests ---- ----


def test_02_aggregate_data_insert(psyscale_db: PsyscaleDB, caplog, AAPL_MIN_DATA):
    # pkey and metadata fetch already asserted working in first test.
    aapl = psyscale_db.search_symbols({"store_minute": True})[0]
    metadata = psyscale_db.stored_metadata(aapl["pkey"], _all=True)[0]

    with pytest.raises(ValueError):
        # Should error since a 'rth' column is needed but not given.
        psyscale_db.upsert_series(aapl["pkey"], metadata, AAPL_MIN_DATA, None)

    with pytest.raises(AttributeError):
        # Should error since a column is missing
        psyscale_db.upsert_series(aapl["pkey"], metadata, AAPL_MIN_DATA.drop(columns="date"), None)

    # These inter trackers should not yet be populated
    assert not hasattr(psyscale_db, "_altered_tables")
    assert not hasattr(psyscale_db, "_altered_tables_mdata")

    # Should Work, renames columns and all (column renaming tests in series_df tests)
    # And Dropps Extra Columns
    with caplog.at_level("DEBUG"):
        psyscale_db.upsert_series(aapl["pkey"], metadata, AAPL_MIN_DATA, "NYSE")

    # Assert there was no cursor error in the upsert function.
    assert all(record.levelname != "ERROR" for record in caplog.records)

    # Now they should be populated
    assert hasattr(psyscale_db, "_altered_tables")
    assert hasattr(psyscale_db, "_altered_tables_mdata")


def test_03_check_inserted_data(psyscale_db: PsyscaleDB, caplog):
    aapl = psyscale_db.search_symbols({"store_minute": True})[0]

    # Data is stored, but the metadata table should not reflect this yet.
    # it should only show this once the refresh data has been called.
    metadata = psyscale_db.stored_metadata(aapl["pkey"])
    assert len(metadata) == 0

    assert hasattr(psyscale_db, "_altered_tables")
    assert hasattr(psyscale_db, "_altered_tables_mdata")

    with caplog.at_level("INFO"):
        psyscale_db.refresh_aggregate_metadata()

    # Assert tracking variables are cleaned up
    assert not hasattr(psyscale_db, "_altered_tables")
    assert not hasattr(psyscale_db, "_altered_tables_mdata")

    # check what's available
    metadata = psyscale_db.stored_metadata(aapl["pkey"])
    assert len(metadata) == 4  # One for inserted data, 3 more for the aggregates

    minute_metadata = [m for m in metadata if m.timeframe == pd.Timedelta("1min")][0]

    # check that the full dataset got inserted
    raw_data = Series_DF(pd.read_csv("example_data/AAPL_1min.csv"), "NYSE")
    raw_data.df.drop(columns=set(raw_data.columns).difference(AGGREGATE_ARGS), inplace=True)
    raw_data.df.set_index(keys=pd.RangeIndex(0, 2083), inplace=True)
    raw_data.df["rth"] = raw_data.df["rth"].astype("int64")

    inserted_data = psyscale_db.get_series(
        aapl["pkey"],
        pd.Timedelta("1min"),
        rth=False,
        rtn_args={"open", "high", "low", "close", "volume", "rth"},
        mdata=minute_metadata,
    )
    assert inserted_data is not None

    for col in raw_data.columns:
        if col == "volume":
            assert_series_equal(raw_data.df[col].astype("int64"), inserted_data[col])
        else:
            assert_series_equal(raw_data.df[col], inserted_data[col])

    # check that only the 'rth' data cna be retrieved
    raw_data.df = raw_data.df[raw_data.df["rth"] == 0]
    raw_data.df.set_index(keys=pd.RangeIndex(0, 780), inplace=True)

    inserted_data = psyscale_db.get_series(
        aapl["pkey"],
        pd.Timedelta("1min"),
        rth=True,
        rtn_args={"open", "high", "low", "close", "volume", "rth"},
    )
    assert inserted_data is not None

    for col in raw_data.columns:
        if col == "volume":
            # very strange quirk, but on retrieval, since data is being read as a csv stream
            # the decimals can be dropped leading to colume becoming an int. can happen on
            # all columns, but really only likely on the volume column.
            # tbh smells like a feature not a bug
            assert_series_equal(raw_data.df[col].astype("int64"), inserted_data[col])
        else:
            assert_series_equal(raw_data.df[col], inserted_data[col])


def test_04_upsert_on_conflict_states(psyscale_db: PsyscaleDB, AAPL_MIN_DATA, caplog):
    aapl = psyscale_db.search_symbols({"store_minute": True})[0]
    metadata = psyscale_db.stored_metadata(aapl["pkey"])[0]

    alt_data = Series_DF(AAPL_MIN_DATA, "NYSE").df
    extra_cols = set(alt_data.columns).difference(AGGREGATE_ARGS)
    alt_data = AAPL_MIN_DATA.drop(columns=extra_cols)
    # Make some Adjustments
    alt_data.loc[0, "low"] = -1
    alt_data.loc[0, "high"] = -1
    alt_data.loc[0, "open"] = -1
    alt_data.loc[0, "close"] = -1
    alt_data.loc[0, "volume"] = -1

    with pytest.raises(DatabaseError):
        psyscale_db.upsert_series(aapl["pkey"], metadata, alt_data, "NYSE", on_conflict="error")

    with caplog.at_level("WARNING"):
        psyscale_db.upsert_series(aapl["pkey"], metadata, alt_data, "NYSE", on_conflict="ignore")
    assert any(r.levelname == "WARNING" for r in caplog.records)
    stored_data = psyscale_db.get_series(aapl["pkey"], pd.Timedelta("1min"), rth=False, limit=10)
    assert stored_data is not None

    assert stored_data["dt"].iloc[0] == alt_data["dt"].iloc[0]
    assert stored_data.iloc[0]["low"] != alt_data.iloc[0]["low"]
    assert stored_data.iloc[0]["high"] != alt_data.iloc[0]["high"]
    assert stored_data.iloc[0]["open"] != alt_data.iloc[0]["open"]
    assert stored_data.iloc[0]["close"] != alt_data.iloc[0]["close"]
    assert stored_data.iloc[0]["volume"] != alt_data.iloc[0]["volume"]

    psyscale_db.upsert_series(aapl["pkey"], metadata, alt_data, "NYSE", on_conflict="update")
    stored_data = psyscale_db.get_series(aapl["pkey"], pd.Timedelta("1min"), rth=False, limit=10)
    assert stored_data is not None

    assert stored_data["dt"].iloc[0] == alt_data["dt"].iloc[0]
    assert_series_equal(stored_data.iloc[0], alt_data.iloc[0])

    # Re-insert & clean the original data for use in following tests
    psyscale_db.upsert_series(aapl["pkey"], metadata, AAPL_MIN_DATA.iloc[0:10], "NYSE", on_conflict="update")
    psyscale_db.refresh_aggregate_metadata()


def test_05_stored_aggregates(psyscale_db: PsyscaleDB, caplog):
    stored_data = psyscale_db.get_series("aapl", pd.Timedelta("30min"), limit=10)
    assert stored_data is not None

    # fmt: off
    expect_df = pd.DataFrame({
        "dt": [
            "2023-05-24 21:00:00+00:00",
            "2023-05-24 21:30:00+00:00",
            "2023-05-24 22:00:00+00:00",
            "2023-05-24 22:30:00+00:00",
            "2023-05-24 23:00:00+00:00",
            "2023-05-24 23:30:00+00:00",
            "2023-05-25 08:00:00+00:00",
            "2023-05-25 08:30:00+00:00",
            "2023-05-25 09:00:00+00:00",
            "2023-05-25 09:30:00+00:00",
        ],
        "open": [171.68, 171.43, 171.61, 171.38, 171.18, 170.88, 171.37, 171.58, 171.72, 171.53],
        "high": [171.69, 171.73, 171.69, 171.40, 171.18, 171.15, 171.80, 171.71, 171.75, 171.68],
        "low": [171.30, 171.32, 171.22, 171.00, 170.85, 170.84, 171.37, 171.41, 171.52, 171.53],
        "close": [171.42, 171.61, 171.31, 171.04, 170.88, 171.00, 171.58, 171.64, 171.54, 171.65],
        "volume": [39688, 42391, 71238, 50434, 34025, 37182, 27388, 9683, 12406, 5810],
    })
    # fmt: on
    expect_df["dt"] = pd.to_datetime(expect_df["dt"])
    assert_frame_equal(stored_data, expect_df)


def test_06_calculated_aggregates(psyscale_db: PsyscaleDB, caplog):
    stored_data = psyscale_db.get_series("aapl", pd.Timedelta("15min"), limit=10)
    assert stored_data is not None

    # fmt: off
    expect_df = pd.DataFrame({
        "dt": [
            "2023-05-24 21:15:00+00:00",
            "2023-05-24 21:30:00+00:00",
            "2023-05-24 21:45:00+00:00",
            "2023-05-24 22:00:00+00:00",
            "2023-05-24 22:15:00+00:00",
            "2023-05-24 22:30:00+00:00",
            "2023-05-24 22:45:00+00:00",
            "2023-05-24 23:00:00+00:00",
            "2023-05-24 23:15:00+00:00",
            "2023-05-24 23:30:00+00:00",
        ],
        "open": [171.68, 171.43, 171.61, 171.61, 171.51, 171.38, 171.14, 171.18, 171.01, 170.88],
        "high": [171.69, 171.70, 171.73, 171.69, 171.62, 171.40, 171.28, 171.18, 171.06, 171.15],
        "low": [171.30, 171.32, 171.36, 171.37, 171.22, 171.00, 171.03, 170.90, 170.85, 170.84],
        "close": [171.42, 171.65, 171.61, 171.63, 171.31, 171.20, 171.04, 171.00, 170.88, 171.14],
        "volume": [39688, 17973, 24418, 37748, 33490, 27235, 23199, 19566, 14459, 13893],
    })
    # fmt: on
    expect_df["dt"] = pd.to_datetime(expect_df["dt"])

    assert_frame_equal(stored_data, expect_df)

    with caplog.at_level("WARNING"):
        stored_data = psyscale_db.get_series("aapl", pd.Timedelta("7.5min"), limit=10)
        assert stored_data is None
    # should raise a warning message that the data cannot be derived from the stored data
    assert any(r.levelname == "WARNING" for r in caplog.records)


# endregion

# region ---- ---- Tick Data Tests ---- ----


def test_07_tick_data_insert(psyscale_db: PsyscaleDB, caplog, TICK_DATA):
    # pkey and metadata fetch already asserted working in first test.
    goog = psyscale_db.search_symbols({"symbol": "goog"})[0]
    metadata = psyscale_db.stored_metadata(goog["pkey"], _all=True)[0]

    with pytest.raises(ValueError):
        # Should error since a 'rth' column is needed but not given.
        psyscale_db.upsert_series(goog["pkey"], metadata, TICK_DATA, None)

    with pytest.raises(AttributeError):
        # Should error since a column is missing
        psyscale_db.upsert_series(goog["pkey"], metadata, TICK_DATA.drop(columns="time"), None)

    # These inter trackers should not yet be populated
    assert not hasattr(psyscale_db, "_altered_tables")
    assert not hasattr(psyscale_db, "_altered_tables_mdata")

    # Should Work, renames columns and all (column renaming tests in series_df tests)
    # And Dropps Extra Columns
    with caplog.at_level("DEBUG"):
        psyscale_db.upsert_series(goog["pkey"], metadata, TICK_DATA, "NYSE")

    # Assert there was no cursor error in the upsert function.
    assert all(record.levelname != "ERROR" for record in caplog.records)

    # Now they should be populated
    assert hasattr(psyscale_db, "_altered_tables")
    assert hasattr(psyscale_db, "_altered_tables_mdata")


def test_08_check_inserted_tick_data(psyscale_db: PsyscaleDB):
    goog = psyscale_db.search_symbols({"store_tick": True})[0]

    # Data is stored, but the metadata table should not reflect this yet.
    # it should only show this once the refresh data has been called.
    metadata = psyscale_db.stored_metadata(goog["pkey"])
    assert len(metadata) == 0

    assert hasattr(psyscale_db, "_altered_tables")
    assert hasattr(psyscale_db, "_altered_tables_mdata")

    psyscale_db.refresh_aggregate_metadata()

    # Assert tracking variables are cleaned up
    assert not hasattr(psyscale_db, "_altered_tables")
    assert not hasattr(psyscale_db, "_altered_tables_mdata")

    # check what's available
    metadata = psyscale_db.stored_metadata(goog["pkey"])
    assert len(metadata) == 3  # One for inserted data, 3 more for the aggregates

    # check that the full dataset got inserted
    raw_data = Series_DF(pd.read_csv("example_data/example_ticks.csv"), "NYSE")
    raw_data.df.drop(columns=set(raw_data.columns).difference(AGGREGATE_ARGS), inplace=True)
    raw_data.df.set_index(keys=pd.RangeIndex(0, 2465), inplace=True)
    raw_data.df["rth"] = raw_data.df["rth"].astype("int64")

    inserted_data = psyscale_db.get_series(
        goog["pkey"],
        pd.Timedelta(0),
        rth=False,
        rtn_args={"volume", "rth", "price"},
    )
    assert inserted_data is not None

    for col in raw_data.columns:
        assert_series_equal(raw_data.df[col], inserted_data[col])


def test_09_stored_tick_aggregates(psyscale_db: PsyscaleDB):
    stored_data = psyscale_db.get_series("goog", pd.Timedelta("15s"), limit=10)
    assert stored_data is not None

    # fmt: off
    expected_df = pd.DataFrame({
        "dt": [
            "2023-05-04 15:15:15+00:00",
            "2023-05-04 15:15:30+00:00",
            "2023-05-04 15:15:45+00:00",
            "2023-05-04 15:16:00+00:00",
            "2023-05-04 15:16:15+00:00",
            "2023-05-04 15:16:30+00:00",
            "2023-05-04 15:16:45+00:00",
            "2023-05-04 15:17:00+00:00",
            "2023-05-04 15:17:15+00:00",
            "2023-05-04 15:17:30+00:00",
        ],
        "open":  [162.56, 162.49, 162.44, 162.29, 162.33, 162.33, 162.24, 162.31, 162.21, 162.45],
        "high":  [162.56, 162.52, 162.46, 162.39, 162.39, 162.37, 162.37, 162.31, 162.46, 162.52],
        "low":   [162.46, 162.44, 162.29, 162.27, 162.29, 162.23, 162.24, 162.21, 162.21, 162.38],
        "close": [162.49, 162.44, 162.29, 162.33, 162.33, 162.24, 162.31, 162.21, 162.45, 162.50],
        "volume": [np.nan] * 10,
    })
    expected_df["dt"] = pd.to_datetime(expected_df["dt"])
    # fmt: on
    assert_frame_equal(stored_data, expected_df)


def test_10_calculated_tick_aggregates(psyscale_db: PsyscaleDB):

    with pytest.raises(ValueError):
        # Cannot due intervals that are not multiples of one second
        stored_data = psyscale_db.get_series("goog", pd.Timedelta("17.5s"), limit=10)

    stored_data = psyscale_db.get_series("goog", pd.Timedelta("7s"), limit=10)
    assert stored_data is not None

    # fmt: off
    df = pd.DataFrame({
        "dt": [
            "2023-05-04 15:15:15+00:00",
            "2023-05-04 15:15:22+00:00",
            "2023-05-04 15:15:29+00:00",
            "2023-05-04 15:15:36+00:00",
            "2023-05-04 15:15:43+00:00",
            "2023-05-04 15:15:50+00:00",
            "2023-05-04 15:15:57+00:00",
            "2023-05-04 15:16:04+00:00",
            "2023-05-04 15:16:11+00:00",
            "2023-05-04 15:16:18+00:00",
        ],
        "open":  [162.56, 162.50, 162.51, 162.52, 162.48, 162.43, 162.29, 162.30, 162.30, 162.32],
        "high":  [162.56, 162.55, 162.52, 162.52, 162.48, 162.44, 162.32, 162.39, 162.35, 162.34],
        "low":   [162.48, 162.46, 162.46, 162.46, 162.42, 162.33, 162.27, 162.30, 162.29, 162.30],
        "close": [162.50, 162.51, 162.52, 162.48, 162.43, 162.33, 162.30, 162.30, 162.32, 162.33],
        "volume": [np.nan] * 10,
    })
    df["dt"] = pd.to_datetime(df["dt"])
    # fmt: on
    assert_frame_equal(stored_data, df)



"""Suppoting Dataclass Objects and Constants for interfacing with a Timescale Database"""

from __future__ import annotations
from enum import StrEnum
import logging
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Literal,
    Optional,
    Protocol,
    Self,
    Tuple,
    get_args,
)

from dataclasses import dataclass, field
from pandas import Timedelta, Timestamp

DEFAULT_ORIGIN_DATE = Timestamp("2000/01/03 00:00", tz="UTC")
DEFAULT_HTF_ORIGIN_DATE = Timestamp("2000/01/01 00:00", tz="UTC")
DEFAULT_AGGREGATES = [
    Timedelta("5min"),
    Timedelta("15min"),
    Timedelta("30min"),
    Timedelta("1h"),
    Timedelta("4h"),
    Timedelta("1D"),
    Timedelta("1W"),
]

log = logging.getLogger("psyscale_log")


def _get_ensured[T](_dict: dict[str, T | None], key: str, default: T) -> T:
    "Enhanced get function to return the default even when the value stored in the dict is None"
    val = _dict.get(key)
    return val if val is not None else default


class Storable(Protocol):
    "Protocol definition for a Class instance that can be dumped into and loaded from JSON"

    @classmethod
    def from_json(cls, json: dict[str, Any]) -> Self:
        "Loads and returns a class instance from a JSON representation of the object"
        raise NotImplementedError

    def to_json(self) -> dict:
        "Returns JSON representation of the object as a dictionary"
        raise NotImplementedError


MetadataArgs = Literal[
    "pkey",
    "table_name",
    "schema_name",
    "start_date",
    "end_date",
    "timeframe",
    "is_raw_data",
    "trading_hours_type",
]
METADATA_ARGS = set(v for v in get_args(MetadataArgs))


@dataclass
class MetadataInfo:
    """
    Construct for returning symbol metadata.

    Dataclass contains the earliest and latest recorded datapoint (start_date & end_date
    respectfully) for a given data table (schema_name & table_name)
    """

    pkey: int
    table_name: str
    schema_name: str | StrEnum
    start_date: Timestamp
    end_date: Timestamp
    table: AssetTable | None = None
    timeframe: Timedelta = field(init=False)

    def __post_init__(self):
        # Allows str/int/datetime args to be passed to constructer and ensure
        # a pd.Timestamp is always stored.
        self.start_date = Timestamp(self.start_date)
        self.end_date = Timestamp(self.end_date)
        if self.table is None:
            self.table = AssetTable.from_table_name(self.table_name)
        self.timeframe = self.table.period


# region -------- -------- -------- Asset Table Dataclass + Functions -------- -------- --------
# pylint: disable='missing-function-docstring'


@dataclass
class AssetTable:
    """
    Dataclass to Facilitate interacting with timeseries data tables

    -- PARAMS --
    - asset_name - The Name of the asset class this belongs to.
    - period - Timedelta of the Aggregation window
    - raw - Whether the data was inserted or calculated.
        * True : Table is raw data inserted from an API
        * False : Table is a Continuous Aggregate from a raw table
    - ext - Whether the asset class contains any ext information
        * True : Contains EXT information at any level of aggregation
        * False : Asset Class only has RTH Data (e.g. Crypto)
    - rth - Whether the table stores RTH data, ETH Data, or Both
        * None: Has 'rth' Boolean Column denoting a mix of RTH & ETH Data
        * True: No 'rth' Column, All data is RTH
        * False: No 'rth' Column, All data is ETH
    - _origin - Bucketing Start time for timeframes < 4 Weeks
    - _origin_htf - Time Bucket Start Time for timeframes >= 4 Weeks
    """

    asset_class: str
    period: Timedelta
    raw: bool
    ext: bool
    rth: Optional[bool]
    _origin_ltf: Optional[Timestamp] = None
    _origin_htf: Optional[Timestamp] = None
    _table_name: str = field(init=False)

    def __post_init__(self):
        """
        Format storage parameters into a series table name
        PARAMS:
        """
        if self.period % Timedelta("1s") != Timedelta(0):
            raise ValueError(
                "Cannot Generate an Asset Table with a period that isn't a"
                f" whole number of seconds. Given : {self.period}",
            )

        self._table_name = self.asset_class + "_" + str(int(self.period.total_seconds()))

        if self._origin_ltf is not None and self._origin_htf is None:
            # Normalize to midnight of the first of whatever month was given.
            self._origin_htf = self._origin_ltf.replace(day=1).normalize()

        if self.raw:
            self._table_name += "_raw"  # Raw Data Inserted from an API

        if self.ext is False:
            return  # Neither RTH and ETH Data
        if self.rth is None:
            self._table_name += "_ext"  # Both RTH and ETH Data
        elif self.rth:
            self._table_name += "_rth"
        else:
            self._table_name += "_eth"

    def __eq__(self, other: Self) -> bool:
        if not isinstance(other, AssetTable):
            raise NotImplementedError("Can only Compare between an AssetTable to an AssetTable")
        # Compares asset_class, period, raw, ext, and rth all in one check
        return self._table_name == other._table_name

    def __hash__(self) -> int:
        "Converts the table name to a concatenated list of ascii integers."
        return int("".join(map(lambda x: str(ord(x)), self._table_name)))

    def __repr__(self) -> str:
        return self._table_name

    @property
    def table_name(self) -> str:
        return self._table_name

    @property
    def has_rth(self) -> bool:
        "Whether or not the table contains a 'rth'::SMALLINT column or not"
        return self.ext and not self.rth

    @property
    def origin_ts(self) -> Timestamp:
        if self.period < Timedelta("4W"):
            return DEFAULT_ORIGIN_DATE if self._origin_ltf is None else self._origin_ltf
        else:
            return self._origin_htf if self._origin_htf is not None else DEFAULT_HTF_ORIGIN_DATE

    @property
    def origin(self) -> str:
        return self.origin_ltf if self.period < Timedelta("4W") else self.origin_htf

    @property
    def origin_ltf(self) -> str:
        return str(DEFAULT_ORIGIN_DATE if self._origin_ltf is None else self._origin_ltf)

    @property
    def origin_htf(self) -> str:
        return str(self._origin_htf if self._origin_htf is not None else DEFAULT_HTF_ORIGIN_DATE)

    @property
    def psql_interval(self) -> str:
        return str(int(self.period.total_seconds())) + " seconds"

    @classmethod
    def from_table_name(cls, table_name: str) -> Self:
        """
        Return an AssetTable with the storage paramaters from a given table name.
        Near Inverse Operation to forming the table name on construction. (Origin is lost)
        """
        if table_name.endswith("_rth"):
            ext, rth = True, True
            table_name = table_name.removesuffix("_rth")
        elif table_name.endswith("_eth"):
            ext, rth = True, False
            table_name = table_name.removesuffix("_eth")
        elif table_name.endswith("_ext"):
            ext, rth = True, None
            table_name = table_name.removesuffix("_ext")
        else:
            ext, rth = False, None

        raw = table_name.endswith("_raw")
        if raw:
            table_name = table_name.removesuffix("_raw")

        parts = table_name.split("_")

        # Pop off the period in seconds
        period = Timedelta(seconds=int(parts.pop()))
        # Rejoin on the delimiter in case the asset_class str included the delimiter.
        asset_class = "_".join(parts)

        return cls(asset_class, period, raw, ext, rth)


# pylint: disable=line-too-long
@dataclass
class TimeseriesConfig:
    """
    Defines the storage scheme of timeseries data within the postgres database.
    3 Of these objects are used by the TimescaleDB interface, One for each of the following Schema:
    Tick Data, Minute Data, Pre-aggregated Data.

    - Tick Data: Generates aggregates from raw Ticks information.
    - Minute Data: Generates Aggregates from Pre-aggrigated Minute Data.
        * From Sources such as Alpaca & Polygon.io
    - Pre-Aggregated: Only Aggregates are stored, Little/No information is derived.
        * From Sources such as Trading-View & yFinance where data at HTF extends further back in time than LTF Data

    -- PARAMS --
    - asset_types: Iterable[str] - Asset types that can be stored in the Schema. These strings will be used
        as dictionary keys for other input parameters.

        How these types are delineated is important and should be done based on market opening time.
        This is because each Asset Type shares a table. When this table is Aggregated into a Higher-Timeframe
        it will be done with a constant Bucketing Origin Timestamp.
        e.g. US_Stock Data has an origin timestamp of 2000/01/03 8:30 EST for RTH and 2000/01/03 04:00 EST for ETH
        While a Forex Aggregate needs an origin timestamp of 2000/01/03 00:00 UTC. If These assets were stored
        in the same table, some of the Higher Timeframe Aggregates would be incorrect.

    - rth_origins: Dict[Asset_type:str, pd.Timestamp]
        : These Timestamps are Origins for Aggregating Regular-Trading-Hours at periods of less than 4 Weeks
        This should be and should be the market opening time on the first trading day of a week.
        * Optional, If None is given then 2000/01/03 00:00:00 UTC is Assumed.
        * 'default' can be passed as a key to override stated default parameter.

    - eth_origins: Dict[Asset_type:str, pd.Timestamp]
        : These Timestamps are Origins for Extended-Trading-Hours.
        * Optional, If None is given then the RTH_Origin is assumed thus ignoring any RTH/ETH Distinction
        * 'default' can be passed as a key to override stated default parameter.

    - htf_origins: Dict[Asset_type:str, pd.Timestamp]
        : These Timestamps are Origins for Aggregation Periods of 4 Weeks or Larger. Should generally align with
        at least midnight of the first of the month, ideally also the first of a year, decade, and millennium.
        * Optional, If None is given then the 2000/01/01 00:00:00 UTC is Assumed.
        * 'default' can be passed as a key to override stated default parameter.

    - calculated_periods: Dict[Asset_type:str, List[pd.Timedelta]]
        : The Aggregate Periods that should be calculated and stored.
        * The Minimum value is 1 second for Tick Data, and 2 Minutes for Minute Data
        * The Default Aggregates are [5min, 15min, 30min, 1h, 4h, 1D, 1W]
        * 'default' can be passed as a key to override stated default parameter.

    - stored_periods: Dict[Asset_type:str, List[pd.Timedelta]]
        : The Aggregate Periods that should will be inserted from an API and Joined on retrieval.
        * There are no Default Inserted Aggregate
        * 'default' can be passed as a key to override stated default parameter.

    - prioritize_rth: Dict[Asset_type:str, Bool | None]
        : Determines the Behavior on Conflict, per asset, when storing RTH and ETH Aggregates.
        If given a differing RTH & ETH Origin Timestamp for an asset, there is a chance for a given aggregation
        period to result in differing period start times. This bool determines what should be stored.
        e.g. NYSE 1h ETH Periods start aggregation at 8AM/9AM while RTH starts aggregating at 8:30AM.

        * True - Only store the RTH Aggregate at HTFs (Calculate ETH at Runtime) (Default Behavior)
        * False - Only store the ETH Aggregate at HTFs (Calculate RTH at Runtime)
        * None - Store Both the RTH and ETH Aggregate at HTFs
        * 'default' can be passed as a key to override stated default parameter.
    """

    asset_classes: Iterable[str]
    rth_origins: dict[str, Timestamp | None] = field(default_factory=dict)
    eth_origins: dict[str, Timestamp | None] = field(default_factory=dict)
    htf_origins: dict[str, Timestamp | None] = field(default_factory=dict)
    prioritize_rth: dict[str, bool | None] = field(default_factory=dict)
    calculated_periods: dict[str, list[Timedelta] | None] = field(default_factory=dict)
    stored_periods: dict[str, list[Timedelta] | None] = field(default_factory=dict)

    def __post_init__(self):
        # Get all the desired Defaults from the given input
        self._std_tables: Dict[str, List[AssetTable]] = {}
        self._rth_tables: Dict[str, List[AssetTable]] = {}
        self._eth_tables: Dict[str, List[AssetTable]] = {}
        self._inserted_tables: Dict[str, List[AssetTable]] = {}

        _default_rth = _get_ensured(self.rth_origins, "default", DEFAULT_ORIGIN_DATE)
        _default_eth = _get_ensured(self.eth_origins, "default", _default_rth)
        _default_htf = _get_ensured(self.htf_origins, "default", DEFAULT_HTF_ORIGIN_DATE)
        _default_aggs = _get_ensured(self.calculated_periods, "default", DEFAULT_AGGREGATES)

        _default_raw_aggs = _get_ensured(self.stored_periods, "default", [])
        _default_priority = self.prioritize_rth.get("default", True)

        for asset_class in self.asset_classes:
            asset_aggregates = _get_ensured(self.calculated_periods, asset_class, _default_aggs)
            inserted_aggregates = _get_ensured(self.stored_periods, asset_class, _default_raw_aggs)

            # Error check to Ensure no overlapping aggregate and inserted timeframes
            # No fuckin way I'm managing that mess of trying to detect what tables need to be JOIN'd
            overlap = {*asset_aggregates}.intersection(inserted_aggregates)
            if len(overlap) > 0:
                raise AttributeError(
                    f"Asset Class {asset_class} is set to store *and* aggregate the timeframes : \n{overlap}\n\n"
                    "Timeframes can only be Stored or Aggregated, not Both."
                )

            asset_rth_origin = _get_ensured(self.rth_origins, asset_class, _default_rth)
            asset_eth_origin = _get_ensured(self.eth_origins, asset_class, _default_eth)
            asset_htf_origin = _get_ensured(self.htf_origins, asset_class, _default_htf)

            # Store all the asset origins so they can be referenced later (to insert into the DB)
            # Ensures the each key is valid and will yield a Timestamp
            self.rth_origins[asset_class] = asset_rth_origin
            self.eth_origins[asset_class] = asset_eth_origin
            self.htf_origins[asset_class] = asset_htf_origin

            asset_args = {
                "asset_class": asset_class,
                "_origin_htf": asset_htf_origin,
            }

            # region ---- ---- Create All Associated Asset Tables ---- ----
            if asset_rth_origin == asset_eth_origin:
                # No EXT Information for this asset table. Simplify Table creation.
                asset_args |= {
                    "_origin_ltf": asset_rth_origin,
                    "ext": False,
                    "rth": None,
                }
                self._std_tables[asset_class] = [
                    AssetTable(period=period, raw=False, **asset_args) for period in asset_aggregates
                ]
                self._inserted_tables[asset_class] = [
                    AssetTable(period=period, raw=True, **asset_args) for period in inserted_aggregates
                ]
                continue

            # Get the ETH/RTH conflicting aggregates and create the appropriate tables
            asset_args |= {"ext": True, "raw": False, "_origin_ltf": asset_eth_origin}
            std, rth_only, eth_only = _determine_conflicting_timedeltas(
                asset_aggregates,
                asset_rth_origin,
                asset_eth_origin,
                self.prioritize_rth.get(asset_class, _default_priority),
            )

            # std_tables (that have an ext column) use eth_origin since it's earlier than rth_origin
            self._std_tables[asset_class] = [AssetTable(period=period, rth=None, **asset_args) for period in std]
            self._eth_tables[asset_class] = [AssetTable(period=period, rth=False, **asset_args) for period in eth_only]

            # Union Operator Overrides existing keys
            asset_args |= {"_origin_ltf": asset_rth_origin}

            self._rth_tables[asset_class] = [AssetTable(period=period, rth=True, **asset_args) for period in rth_only]

            asset_args |= {"raw": True}

            std, rth_only, eth_only = _determine_conflicting_timedeltas(
                inserted_aggregates,  # re-determine conflicts for the inserted aggs.
                asset_rth_origin,
                asset_eth_origin,
                self.prioritize_rth.get(asset_class, _default_priority),
            )

            self._inserted_tables[asset_class] = [
                AssetTable(period=period, rth=True, **asset_args)
                for period in inserted_aggregates
                if period in rth_only
            ]

            asset_args |= {"_origin_ltf": asset_eth_origin}

            self._inserted_tables[asset_class].extend(
                AssetTable(period=period, rth=False, **asset_args)
                for period in inserted_aggregates
                if period in eth_only
            )
            self._inserted_tables[asset_class].extend(
                AssetTable(period=period, rth=None, **asset_args) for period in inserted_aggregates if period in std
            )
            # endregion

    def _ext_important(self, asset_class: str) -> bool:
        # private_method, Assume Asset_class is known
        return self.eth_origins[asset_class] != self.rth_origins[asset_class]

    def all_tables(self, asset_class: str, *, include_raw: bool = True) -> List[AssetTable]:
        if asset_class not in self.asset_classes:
            raise KeyError(f"{asset_class = } is not a known asset type.")

        base_list = self._inserted_tables.get(asset_class, []) if include_raw else []
        return (
            base_list
            + self._std_tables.get(asset_class, [])
            + self._rth_tables.get(asset_class, [])
            + self._eth_tables.get(asset_class, [])
        )

    def std_tables(self, asset_class: str, inserted: bool = False) -> List[AssetTable]:
        if asset_class not in self.asset_classes:
            raise KeyError(f"{asset_class = } is not a known asset type.")

        raws = self.raw_tables(asset_class, "std") if inserted else []
        if asset_class not in self._std_tables:
            return raws

        if not inserted:
            return self._std_tables[asset_class]
        else:
            return self._std_tables[asset_class] + raws

    def rth_tables(self, asset_class: str, inserted: bool = False) -> List[AssetTable]:
        if asset_class not in self.asset_classes:
            raise KeyError(f"{asset_class = } is not a known asset type.")

        raws = self.raw_tables(asset_class, "rth") if inserted else []
        if asset_class not in self._rth_tables:
            return raws

        if not inserted:
            return self._rth_tables[asset_class]
        else:
            return self._rth_tables[asset_class] + raws

    def eth_tables(self, asset_class: str, inserted: bool = False) -> List[AssetTable]:
        if asset_class not in self.asset_classes:
            raise KeyError(f"{asset_class = } is not a known asset type.")

        raws = self.raw_tables(asset_class, "eth") if inserted else []
        if asset_class not in self._eth_tables:
            return raws

        if not inserted:
            return self._eth_tables[asset_class]
        else:
            return self._eth_tables[asset_class] + raws

    def raw_tables(self, asset_class: str, rth: Literal["all", "std", "rth", "eth"] = "all") -> List[AssetTable]:
        if asset_class not in self.asset_classes:
            raise KeyError(f"{asset_class = } is not a known asset type.")
        if asset_class not in self._inserted_tables:
            return []

        if rth == "all":
            return self._inserted_tables[asset_class]
        elif rth == "std":
            return [tbl for tbl in self._inserted_tables[asset_class] if tbl.rth is None]
        elif rth == "rth":
            return [tbl for tbl in self._inserted_tables[asset_class] if tbl.rth is True]
        else:
            return [tbl for tbl in self._inserted_tables[asset_class] if tbl.rth is False]

    def get_aggregation_source(self, desired_table: AssetTable, rtn_self: bool = False) -> AssetTable:
        """
        Given the desired AssetTable, return the most appropriate AssetTable to pull or derive
        the desired data from. When rtn_self == True this function can return the table passed
        if it is in the table config
        """
        # Handle the case when requesting tick data separately.
        # Timedelta == 0 interferes with the modulus operation
        if desired_table.period == Timedelta(0):
            tbls = self.raw_tables(desired_table.asset_class)
            divisor_tbls = [tbl for tbl in tbls if tbl.period == Timedelta(0)]
            if len(divisor_tbls) == 0:
                raise AttributeError("Requesting Tick Data from an Asset Class that has no Tick data")
        else:
            tbls = self.all_tables(desired_table.asset_class)

            # All Tables that are an even period divider of what is desired
            divisor_tbls = [
                tbl for tbl in tbls if tbl.period == Timedelta(0) or desired_table.period % tbl.period == Timedelta(0)
            ]

            if len(divisor_tbls) == 0:
                raise AttributeError(
                    f"Desired Table's Timeframe, {desired_table.period}, "
                    "cannot be derived from info in the database."
                )

        if not rtn_self:
            # Ensure the desired table is removed from the pool of possible returns
            divisor_tbls = [tbl for tbl in divisor_tbls if tbl != desired_table]

        # Ext information does not matter in this asset_class, return highest timeframe table
        if not self._ext_important(desired_table.asset_class):
            divisor_tbls.sort(key=lambda x: x.period)
            return divisor_tbls[-1]

        # Match EXT Information and then return the highest timeframe table
        if desired_table.rth is None:
            # rth == None must aggregate from rth == None
            ext_matches = [tbl for tbl in divisor_tbls if tbl.rth is None]
        elif desired_table.rth is True:
            # rth == True aggregates to rth timeframe, rth == None / False aggregate to 'eth timeframe'
            # => When 'rth' True, we must aggregate from a raw table when 'ext' matters
            ext_matches = [tbl for tbl in divisor_tbls if tbl.rth or (tbl.raw and tbl.rth is None)]
        else:
            ext_matches = [tbl for tbl in divisor_tbls if tbl.rth is None or tbl.rth == desired_table.rth]

        if len(ext_matches) == 0:
            raise AttributeError(
                "Desired Table's RTH State & timeframe combination cannot be derived from database. "
                f" Desired Table = {desired_table}"
            )

        ext_matches.sort(key=lambda x: x.period)
        return ext_matches[-1]

    def get_selection_source_table(self, desired_table: AssetTable) -> Tuple[AssetTable, bool]:
        """
        Given the desired AssetTable, return the most appropriate AssetTable to pull or derive
        the desired data from. The Returned bool denotes when data must be selected or derived

        When True, the data is stored in the returned AssetTable and can be selected.
        when False, the desired data must be aggregated from the returned AssetTable.

        The desired table should, in the vast majority of cases, have it's RTH property set to
        True or False, If set to None, it is likely to unnecessarily aggregate higher
        timeframe data that may be stored and can be selected instead.
        As a consequence, the desired_table's ext parameter is unused.
        """
        tbl = self.get_aggregation_source(desired_table, rtn_self=True)
        return tbl, tbl.period != desired_table.period

    def get_tables_to_refresh(self, altered_table: AssetTable) -> List[AssetTable]:
        "Return all the tables that need to be refreshed for a given table that has been altered"
        all_aggs = self.all_tables(altered_table.asset_class, include_raw=False)
        # Really this is pretty inefficient and will cause more updates than needed, but I mean,
        # Does it really matter given data will probably be inserted once a day at most? No.
        filtered_aggs = [agg for agg in all_aggs if agg.period > altered_table.period]
        filtered_aggs.sort(key=lambda x: x.period)  # Ensure aggregates are ordered.
        return filtered_aggs

    @classmethod
    def from_table_names(
        cls,
        names: list[str],
        origins: dict[str, Tuple[Timestamp, Timestamp, Timestamp]] = {},
    ):
        """
        Reconstruct the Timeseries Config from the given asset names.
        Origin Timestamps must be given in a dictionary of
        [asset_class:str, (rth_origin, eth_origin, htf_origin)]
        """
        filtered_names = [name for name in names if name.startswith("_")]
        if len(filtered_names) > 0:
            log.debug(
                "Filtered out table names %s when reconstructing TimeseriesConfig from Database table names.",
                filtered_names,
            )

        tables = []
        for name in names:
            try:
                tables.append(AssetTable.from_table_name(name))
            except ValueError:
                log.warning("Timeseries Database contains an invalid table name: %s", name)

        tables.sort(key=lambda tbl: tbl.period)
        asset_classes = {table.asset_class for table in tables}

        cls_inst = cls(asset_classes)  # type:ignore

        # Reconstruct the std, raw, eth, and inserted table dictionaries.
        for asset_class in asset_classes:
            # Get all the tables for this asset class
            cls_tables = [table for table in tables if table.asset_class == asset_class]

            if asset_class in origins:
                cls_inst.rth_origins[asset_class] = origins[asset_class][0]
                cls_inst.eth_origins[asset_class] = origins[asset_class][1]
                cls_inst.htf_origins[asset_class] = origins[asset_class][2]

            # Store the raw inserted tables
            cls_inst._inserted_tables[asset_class] = [table for table in cls_tables if table.raw]
            cls_inst.stored_periods[asset_class] = [table.period for table in cls_inst._inserted_tables[asset_class]]

            # Remove the Raw Tables so that doesn't need to be checked on following generators
            cls_tables = [table for table in cls_tables if not table.raw]

            cls_inst._std_tables[asset_class] = [table for table in cls_tables if table.rth is None]
            cls_inst._rth_tables[asset_class] = [table for table in cls_tables if table.rth is True]
            cls_inst._eth_tables[asset_class] = [table for table in cls_tables if table.rth is False]
            cls_inst.calculated_periods[asset_class] = [table.period for table in cls_tables]

        return cls_inst


def _determine_conflicting_timedeltas(
    periods: List[Timedelta],
    rth_origin: Timestamp,
    eth_origin: Timestamp,
    prioritize_rth: Optional[bool],
) -> Tuple[List[Timedelta], List[Timedelta], List[Timedelta]]:
    """
    Determines if an aggregate period would conflict when aggregating on RTH and ETH Times.
    Returns the sorted time periods that need to be stored as a tuple of lists:
    ([Combined ETH/RTH],[RTH only times],[ETH only times])
    """
    # rth_origin and eth_origin are guerenteed to be different at this stage.

    eth_delta = rth_origin - eth_origin
    # These checks are currently in place since it is unclear if these edge-cases would cause problems
    # Safer to assume they will and remove this limitation later on if it is needed and doesn't cause any issues.
    if eth_delta < Timedelta(0):
        raise ValueError("TimescaleDB ETH Aggregate Origin must occur before RTH Origin.")
    if eth_delta >= Timedelta("1D"):
        raise ValueError("TimescaleDB RTH Origin Must be less than 1D after ETH Origin.")

    periods.sort()
    std_periods, rth_periods, eth_periods = [], [], []

    for period in periods:
        if period == Timedelta(0):
            std_periods.append(period)
            continue

        remainder = eth_delta % period
        if remainder == Timedelta(0):
            std_periods.append(period)
        elif prioritize_rth:
            rth_periods.append(period)
        elif prioritize_rth is None:
            rth_periods.append(period)
            eth_periods.append(period)
        else:
            eth_periods.append(period)

    return std_periods, rth_periods, eth_periods


# endregion



"""An interface for reading and commiting data to a database"""

from dataclasses import dataclass, field
from enum import StrEnum
import logging
import os
import subprocess
from pathlib import Path
from inspect import stack
from contextlib import contextmanager
from typing import (
    Callable,
    Dict,
    Iterator,
    List,
    Literal,
    Mapping,
    Optional,
    Self,
    Tuple,
    TypeAlias,
    overload,
)
from urllib.parse import parse_qs, quote_plus, unquote, urlparse

import psycopg as pg
import psycopg.rows as pg_rows
from psycopg import OperationalError, sql
from psycopg.pq._enums import ExecStatus
from psycopg_pool import ConnectionPool, PoolTimeout

from .psql import (
    GenericTbls,
    Operation as Op,
    OperationMap,
    Schema,
    AssetTbls,
    Commands,
)


# region ----------- Database Structures  -----------

log = logging.getLogger("psyscale_log")

# Get the Timescale.yml in the folder this file is stored in.
DEFAULT_YML_PATH = Path(__file__).parent.joinpath("timescale.yml").as_posix()
TIMESCALE_IMAGE = "timescale/timescaledb-ha:pg17"
POOL_GEN_TIMEOUT = 5  # seconds to wait for the connection pool to be generated
LOCAL_POOL_GEN_TIMEOUT = 0.2  # wait time when starting a local Connection Pool
# endregion
# pylint: disable='protected-access'

DictCursor: TypeAlias = pg.Cursor[pg_rows.DictRow]
TupleCursor: TypeAlias = pg.Cursor[pg_rows.TupleRow]

BASE_ENV_VARS = [
    "PSYSCALE_HOST",
    "PSYSCALE_PORT",
    "PSYSCALE_USER",
    "PSYSCALE_DB_NAME",
    "PSYSCALE_PASSWORD",
]
YML_ENV_VARS = [
    "PSYSCALE_PORT",
    "PSYSCALE_USER",
    "PSYSCALE_DB_NAME",
    "PSYSCALE_PASSWORD",
    "PSYSCALE_VOLUME_PATH",
]


@dataclass
class PsyscaleConnectParams:
    "Dataclass to derive and store Postgres Database Connection Parameters"

    host: str
    port: int
    user: str
    password: str
    database: str
    sslmode: Optional[str] = None
    application_name: Optional[str] = None
    volume_path: Optional[str] = None
    env_init: bool = field(default=False, init=False)
    url: str = field(init=False)

    @property
    def is_local(self) -> bool:
        "Return true when connection points to a local host"
        return self.host in {"localhost", "127.0.0.1", "::1"}

    def __post_init__(self):
        "Format Params into formatted URL."
        user_info = ""
        if self.user:
            user_info += quote_plus(self.user)
            if self.password:
                user_info += f":{quote_plus(self.password)}"
            user_info += "@"

        query_params = []
        if self.sslmode:
            query_params.append(f"sslmode={quote_plus(self.sslmode)}")
        if self.application_name:
            query_params.append(f"application_name={quote_plus(self.application_name)}")
        query_str = f"?{'&'.join(query_params)}" if query_params else ""

        self.url = f"postgresql://{user_info}{self.host}:{self.port}/{self.database}{query_str}"

    @classmethod
    def from_url(cls, url: str) -> Self:
        "Parses a PostgreSQL connection URL into a PsyscaleConnectParams instance."
        parsed = urlparse(url)
        query = parse_qs(parsed.query)

        if parsed.scheme not in ("postgres", "postgresql"):
            raise ValueError(f"Invalid URL scheme '{parsed.scheme}'. Expected 'postgres' or 'postgresql'.")

        return cls(
            host=parsed.hostname or "localhost",
            port=parsed.port or 5432,
            user=unquote(parsed.username) if parsed.username else "",
            password=unquote(parsed.password) if parsed.password else "",
            database=parsed.path.lstrip("/") if parsed.path else "",
            sslmode=query.get("sslmode", [None])[0],
            application_name=query.get("application_name", [None])[0],
        )

    @classmethod
    def from_env(cls) -> Self:
        """
        Return a PsyscaleConnectParams instance from environment variables parameters.
        Note: This function does not search for and load env variables from a .env, it only
        tries to pull the variables from the currently loaded environment variables.
        """
        if url := os.getenv("PSYSCALE_URL"):
            inst = cls.from_url(url)
        elif any(map(os.getenv, BASE_ENV_VARS)):
            # Ensure at least one ENV was given before supplying defaults
            inst = cls(
                host=os.getenv("PSYSCALE_HOST") or "localhost",
                port=int(os.getenv("PSYSCALE_PORT") or 5432),
                user=os.getenv("PSYSCALE_USER") or "postgres",
                database=os.getenv("PSYSCALE_DB_NAME") or "postgres",
                password=os.getenv("PSYSCALE_PASSWORD") or "password",
                sslmode=os.getenv("PSYSCALE_SSLMODE"),
                application_name=os.getenv("PSYSCALE_APP_NAME"),
            )
            # Can only docker_compose yml init when all these vars are present
            inst.env_init = all(map(os.getenv, YML_ENV_VARS))
        else:
            raise AttributeError("Cannot Initialize PsyscaleDB from ENV. No Env variables given.")

        if inst.is_local:
            # Get additional params if using a local database
            inst.volume_path = os.getenv("PSYSCALE_VOLUME_PATH")

        return inst


class PsyscaleCore:
    """
    Core Synchronous client interface for connecting to a PostgreSQL Database.

    This class contains all the necessary functionality needed to interact
    with the Database at runtime. If provided environment variables that point to
    a local host, the initializer will start/create the docker container as needed.

    Additional functionality (Such as one-off configuration scripts, Data Insertion, etc.)
    are handled by the Psyscale Partial Classes. This is done to organize the
    exceedingly large amount of functionality that is needed to manage a Database.
    """

    def __init__(
        self,
        conn_params: Optional[PsyscaleConnectParams | str] = None,
        *,
        down_on_del: bool = False,
        docker_compose_fpath: Optional[str] = None,
    ):
        """
        Initilize the PsyscaleDB Client. When the Class detects that the connection parameters given
        point to a local database the Client will attempt to start the docker container when needed
        using the sub-process standard library to execute docker commands.

        If this is the first time the database is being run it will initilize the needed docker
        container using the included docker_compose yml. This includes generating the desired
        mounting directory.

        -- PARAMS --
        - conn_params : Optional[PsyscaleConnectParams | str]
            - A string argument is interpreted as a formatted connection url.
            - When None is given the Client will initialize the database using environment variables

        - docker_compose_fpath: Optional[str]
            - Optional File path to point to a custom docker compose yml. Will only be used if the
            conn_params point to a local database *and* the client cannot immediately connect.
            In that case, this file path will be used when calling 'docker-compose up' using a
            sub-process.

        - down_on_del : Boolean Default = False
            - When True, on delete, this client will call 'docker compose down' using the stored
            docker_compose yml. While this does close out an unneeded docker container, this is not
            advised since it will blindly close the container even if another instance or program
            has a connection to the database.
            - Only useful when pointing to a local database.
            - Assumes that the given connection parameters point to the database that was started
            from the stored yml (either the default or the one passed as an argument)

        """
        self.down_on_del = down_on_del
        if isinstance(conn_params, str):
            conn_params = PsyscaleConnectParams.from_url(conn_params)
        if conn_params is None:
            conn_params = PsyscaleConnectParams.from_env()

        _timeout = LOCAL_POOL_GEN_TIMEOUT if conn_params.is_local else POOL_GEN_TIMEOUT

        try:
            self._pool = ConnectionPool(conn_params.url, open=False, timeout=_timeout)
            self._pool.open(timeout=_timeout)
            log.debug("Health_check: %s", "good" if self._health_check() else "bad")
        except PoolTimeout as e:
            if not (conn_params.is_local and conn_params.env_init):
                raise e

            # Try and start the local database, give extra buffer on the timeout.
            self._init_and_start_localdb(docker_compose_fpath, conn_params.volume_path)
            with self._pool.connection(timeout=2.5) as conn:
                conn._check_connection_ok()

        except OperationalError as e:
            raise e  # Give some more informative info here?

        self.cmds = Commands()
        self.db_cfg = conn_params
        self._ensure_std_schemas_exist()

    def __getitem__(self, args: Tuple[Op, StrEnum]) -> Callable[..., sql.Composed]:
        "Accessor forwarder for the self.cmds object"
        return self.cmds[args]

    def __del__(self):
        if self.down_on_del and self.db_cfg.is_local:
            subprocess.run(
                ["docker-compose", "-f", self.yml_path, "down"],
                capture_output=False,
                check=False,
            )

    def merge_operations(self, _map: OperationMap):
        "Merge additional operations into the Database's stored SQL Command Map"
        self.cmds.merge_operations(_map)

    # region ----------- Connection & Cursor Methods -----------

    # region -- Cursor & Execute Overloading ---

    @overload
    @contextmanager
    def _cursor(
        self,
        dict_cursor: Literal[True],
        *,
        binary: bool = False,
        pipeline: bool = False,
        raise_err: bool = False,
        auto_commit: bool = False,
    ) -> Iterator[DictCursor]: ...
    @overload
    @contextmanager
    def _cursor(
        self,
        dict_cursor: Literal[False] = False,
        *,
        binary: bool = False,
        pipeline: bool = False,
        raise_err: bool = False,
        auto_commit: bool = False,
    ) -> Iterator[TupleCursor]: ...
    @overload
    @contextmanager
    def _cursor(
        self,
        dict_cursor: bool = False,
        *,
        binary: bool = False,
        pipeline: bool = False,
        raise_err: bool = False,
        auto_commit: bool = False,
    ) -> Iterator[TupleCursor]: ...

    @overload
    def execute(
        self,
        cmd: sql.Composed,
        exec_args: Optional[Mapping[str, int | float | str | None]] = None,
        dict_cursor: Literal[False] = False,
    ) -> Tuple[List[Tuple], str | None]: ...
    @overload
    def execute(
        self,
        cmd: sql.Composed,
        exec_args: Optional[Mapping[str, int | float | str | None]] = None,
        dict_cursor: Literal[True] = True,
    ) -> Tuple[List[Dict], str | None]: ...

    # endregion

    @contextmanager
    def _cursor(
        self,
        dict_cursor: bool = False,
        *,
        binary: bool = False,
        pipeline: bool = False,
        raise_err: bool = False,
        auto_commit: bool = False,
    ) -> Iterator[TupleCursor] | Iterator[DictCursor]:
        """
        Returns a cursor to execute commands in a database.

        Default return product is a list of tuples. Returns can be made into lists of dictionaries
        by settign dict_cursor=True. This is less performant for large datasets though.

        Auto_Commit Allows for commands to be done outside of a transaction block which may be
        required for some commands.

        Pipeline is a feature of a cursor, that when set to True, avoids waiting for responses
        before executing new commands. In theory that should increase performance. In practice
        it seemed to only reduce performance.

        raise_err = True Raises any database errors after rolling back the database.
        raise_err = False Rollsback changes, logs an error, then silences the error.
        """
        cursor_factory = pg_rows.dict_row if dict_cursor else pg_rows.tuple_row
        conn: pg.Connection = self._pool.getconn()

        if auto_commit:
            conn.set_autocommit(True)

        try:
            if pipeline:
                with (
                    conn.pipeline(),
                    conn.cursor(row_factory=cursor_factory, binary=binary) as cursor,
                ):
                    yield cursor  # type:ignore : Silence the Dict/Tuple overloading Error
            else:
                with (
                    conn,
                    conn.cursor(row_factory=cursor_factory, binary=binary) as cursor,
                ):
                    yield cursor  # type:ignore : Silence the Dict/Tuple overloading Error
        except pg.DatabaseError as e:
            conn.rollback()  # Reset Database, InFailedSqlTransaction Err thrown if not reset
            log.error("Caught Database Error: \n '%s' \n...Rolling back changes.", e)

            if raise_err:
                raise e  # log the message and continue to let the error bubble
        finally:
            if auto_commit:
                conn.set_autocommit(False)
            else:
                conn.commit()

            self._pool.putconn(conn)

    def execute(
        self,
        cmd: sql.Composed | sql.SQL,
        exec_args: Optional[Mapping[str, int | float | str | None]] = None,
        dict_cursor: bool = False,
    ) -> Tuple[List[Dict] | List[Tuple], str | None]:
        """
        Execution Method to manually invoke a command within the database.

        -- PARAMS --
        - cmd : sql.Composed
            - Composed SQL function using psycopg.sql sub-module
        - exec_args : Optional Mapping
            - When supplied, exec_args will be passed to the cursor and used to populate
            any placeholder args within the formatted command.
        - dict_cursor : boolean
            - When true will return any results as a list of dicts where each item in
            the list is a row of the returned table.
            - When false a list of Tuples per row is returned.

        -- RETURNS --
        Tuple[ List[] , str ] ==> Tuple of query results (if any) and the cursor status
        message returned as a string.
        """
        with self._cursor(dict_cursor) as cursor:
            try:
                log.debug("Executing PSQL Command: %s", cmd.as_string(cursor))
                cursor.execute(cmd, exec_args)

            except pg.DatabaseError as e:
                log.error(
                    "Cursor Execute Exception (%s) occured in '%s' \n  Exception Msg: %s",
                    e.__class__.__name__,
                    stack()[1].function,
                    e,
                )
                return [], cursor.statusmessage

            response = []
            pgr = cursor.pgresult
            # Really sad I have to dig to check if there is data available.
            if pgr is not None and pgr.status == ExecStatus.TUPLES_OK:
                response = cursor.fetchall()

            return response, cursor.statusmessage

    def _health_check(self) -> bool:
        "Simple Ping to the Database to ensure it is alive"
        with self._cursor() as cursor:
            cursor.execute("SELECT 1")
            rsp = cursor.fetchall()
            return rsp[0][0] == 1
        return False

    # endregion

    # region ----------- Private Database Interaction Methods -----------

    def _init_and_start_localdb(self, docker_compose_fpath: Optional[str], vol_path: str | None):
        "Starts Up, via subprocess terminal cmds, a local Docker Container that runs TimescaleDB"
        try:  # Ensure Docker is installed
            subprocess.run(["docker", "--version"], capture_output=True, check=True)
        except subprocess.CalledProcessError as e:
            raise OSError("Cannot Initialize Local PsyscaleDB, OS does not have docker installed.") from e

        # Ensure Timescale Image has been pulled.
        try:
            p = subprocess.run(
                ["docker", "images", TIMESCALE_IMAGE],
                capture_output=True,
                check=True,
            )
        except subprocess.CalledProcessError as e:
            raise OSError("Failed to Read Installed Docker Images, Ensure Docker Engine is Running") from e

        # The following check may only work on windows...
        if len(p.stdout.decode().strip().split("\n")) <= 1:
            # i.e. STDOut only returned table heading and no rows listing available images.
            log.warning(
                "Missing required TimescaleDB Image. Pulling Docker Image: %s",
                TIMESCALE_IMAGE,
            )
            try:
                p = subprocess.run(
                    ["docker", "pull", TIMESCALE_IMAGE],
                    capture_output=True,
                    check=True,
                )
            except subprocess.CalledProcessError as e:
                raise OSError("Could not pull Docker Image.") from e
            log.info("Successfully pulled Docker Image")

        if vol_path and not os.path.exists(vol_path):
            log.info("Making Database Volume Folder at : %s", vol_path)
            os.mkdir(vol_path)

        if docker_compose_fpath is not None:
            # Overwrite Default YML path if given a valid filepath
            if not (os.path.isfile(docker_compose_fpath) and docker_compose_fpath.lower().endswith((".yaml", ".yml"))):
                raise ValueError(f"{docker_compose_fpath = } must be a .yaml/.yml File")
            self.yml_path = docker_compose_fpath
        else:
            # Use Default Docker_Compose Config
            self.yml_path = DEFAULT_YML_PATH

        p = subprocess.run(  # Unfortunately this is the slowest command @ around 0.4s
            ["docker-compose", "-f", self.yml_path, "up", "-d"],
            capture_output=True,
            check=False,
        )

        if p.returncode != 0:
            raise OSError(f"Failed to start Docker-Compose with Err Msg: {p.stderr.decode()}")

    def _ensure_std_schemas_exist(self):
        with self._cursor() as cursor:
            cursor.execute(self[Op.SELECT, GenericTbls.SCHEMA]())
            schemas: set[str] = {rsp[0] for rsp in cursor.fetchall()}

            for schema in {v for v in Schema}.difference(schemas):
                log.info("Creating Schema %s", schema)
                cursor.execute(self[Op.CREATE, GenericTbls.SCHEMA](schema))

    def _get_pkey(self, symbol: str | int) -> int | None:
        if isinstance(symbol, int):
            return symbol

        rsp, _ = self.execute(
            self[Op.SELECT, GenericTbls.TABLE](
                Schema.SECURITY,
                AssetTbls.SYMBOLS,
                [
                    "pkey",
                    "source",
                    "exchange",
                    "asset_class",
                ],
                filters=("symbol", "ILIKE", symbol),
                _limit=3,
            )
        )

        if len(rsp) > 1:
            log.warning(
                "Attempting to fetch or update symbol %s but database contains"
                " multiple of these symbols. Using first match %s",
                symbol,
                rsp[0],
            )

        return None if len(rsp) == 0 else rsp[0][0]

    # endregion



"""Formatted SQL Commands for all Tables & Views Under the 'Security' Schema"""

from json import dumps
from typing import (
    Any,
    Literal,
    Optional,
    get_args,
)

from psycopg import sql

from .enum import Schema, AssetTbls
from .generic import Filter, where, arg_list, limit, select

# pylint: disable=missing-function-docstring, protected-access

# region -------- -------- Security Commands -------- --------

SymbolArgs = Literal[
    "pkey",
    "symbol",
    "name",
    "source",
    "exchange",
    "asset_class",
    "store",
    "store_tick",
    "store_minute",
    "store_aggregate",
]
SYMBOL_ARGS = set(v for v in get_args(SymbolArgs))

STRICT_SYMBOL_ARGS = {
    "pkey",
    "source",
    "exchange",
    "asset_class",
    "store",
    "store_tick",
    "store_minute",
    "store_aggregate",
}


def create_search_functions() -> sql.Composed:
    return sql.SQL("CREATE EXTENSION IF NOT EXISTS pg_trgm;").format()


def create_symbol_table() -> sql.Composed:
    return sql.SQL(
        """
        CREATE TABLE {schema_name}.{table_name}(
            pkey SERIAL PRIMARY KEY,
            symbol TEXT NOT NULL,
            source TEXT NOT NULL,
            exchange TEXT NOT NULL,
            asset_class TEXT NOT NULL,
            name TEXT NOT NULL,
            store_tick BOOLEAN NOT NULL DEFAULT False,
            store_minute BOOLEAN NOT NULL DEFAULT False,
            store_aggregate BOOLEAN NOT NULL DEFAULT False,
            store BOOLEAN GENERATED ALWAYS AS (store_tick OR store_minute OR store_aggregate) STORED,
            attrs jsonb,
            CONSTRAINT unique_asset UNIQUE (symbol, source, exchange),
            CONSTRAINT unique_data_schema CHECK (store_tick::int + store_minute::int + store_aggregate::int <= 1)
        );
        CREATE INDEX attrs_gin_idx ON {schema_name}.{table_name} USING gin (attrs);
        CREATE INDEX name_trgm_idx ON {schema_name}.{table_name} USING gin (name gin_trgm_ops);
        CREATE INDEX symbol_trgm_idx ON {schema_name}.{table_name} USING gin (symbol gin_trgm_ops);
    """
    ).format(
        schema_name=sql.Identifier(Schema.SECURITY),
        table_name=sql.Identifier(AssetTbls.SYMBOLS),
    )


def select_symbols(
    name: Optional[str],
    symbol: Optional[str],
    filters: list[Filter],
    include_attrs: bool = False,
    attrs: Optional[dict[str, Any]] = None,
    _limit: Optional[str | int] = None,
) -> sql.Composed:
    # Don't return the '_score' regardless if attrs is returned or not
    rtn_args = [v for v in get_args(SymbolArgs)]
    if include_attrs:
        rtn_args.append("attrs")
    return sql.SQL(
        """
        WITH _base_matches AS ( 
            SELECT * from {schema_name}.{table_name}{_filters}
        ),
        _graded_matches AS (
            SELECT *, {_score} AS _score FROM {_inner_select}
        )
        SELECT {_rtn_args} FROM _graded_matches WHERE _score > 0 ORDER BY _score DESC {_limit};
    """
    ).format(
        schema_name=sql.Identifier(Schema.SECURITY),
        table_name=sql.Identifier(AssetTbls.SYMBOLS),
        _filters=where(filters),
        _inner_select=_inner_attrs_select(attrs),
        _rtn_args=arg_list(rtn_args),
        _score=_symbol_score(name, symbol),
        _limit=limit(_limit),
    )


def _symbol_score(name: Optional[str], symbol: Optional[str]) -> sql.Composable:
    # Utilizes similarity function from pg_trgm to rank matches by relevancy
    match name, symbol:
        case str(), str():
            stmt = sql.SQL("(similarity(name, {_name}) + similarity(symbol, {_symbol}))")
        case None, str():
            stmt = sql.SQL("similarity(symbol, {_symbol})")
        case str(), None:
            stmt = sql.SQL("similarity(name, {_name})")
        case _:
            stmt = sql.SQL("1")

    return stmt.format(_name=sql.Literal(name), _symbol=sql.Literal(symbol))


def _inner_attrs_select(attrs: Optional[dict[str, Any]] = None) -> sql.Composable:
    if attrs is None:
        # No Additonal Select Statement Necessary
        return sql.SQL("_base_matches")
    else:
        # Perform an Inner Select on _base_matches to test for the given Attrs.
        return sql.SQL("( SELECT * FROM _base_matches WHERE attrs @> {json} )").format(json=dumps(attrs))


def create_symbol_buffer() -> sql.Composed:
    "Temp table in injest symbols from. The Source arg is not present since it is assumed constant."
    return sql.SQL(
        """
        CREATE TEMP TABLE _symbol_buffer (
            symbol TEXT NOT NULL,
            exchange TEXT NOT NULL,
            asset_class TEXT NOT NULL,
            name TEXT NOT NULL,
            attrs jsonb
        ) ON COMMIT DROP;
    """
    ).format(
        schema_name=sql.Identifier(Schema.SECURITY),
        table_name=sql.Identifier(AssetTbls.SYMBOLS),
    )


def copy_symbols(args: list[str]) -> sql.Composed:
    return sql.SQL("COPY _symbol_buffer ({args}) FROM STDIN;").format(
        args=sql.SQL(",").join([sql.Identifier(arg) for arg in args]),
    )


def insert_copied_symbols(source: str) -> sql.Composed:
    return sql.SQL(
        """
        INSERT INTO {schema_name}.{table_name} (source, symbol, name, exchange, asset_class, attrs) 
        SELECT {source}, symbol, name, exchange, asset_class, attrs FROM _symbol_buffer
        ON CONFLICT (symbol, source, exchange) DO NOTHING
        RETURNING symbol;
    """
    ).format(
        schema_name=sql.Identifier(Schema.SECURITY),
        table_name=sql.Identifier(AssetTbls.SYMBOLS),
        source=sql.Literal(source),
    )


def upsert_copied_symbols(source: str) -> sql.Composed:
    return sql.SQL(
        """
        INSERT INTO {schema_name}.{table_name} (source, symbol, name, exchange, asset_class, attrs) 
        SELECT {source}, symbol, name, exchange, asset_class, attrs FROM _symbol_buffer
        ON CONFLICT (symbol, source, exchange)  DO UPDATE
        SET name = EXCLUDED.name, 
            asset_class = EXCLUDED.asset_class,
            attrs = EXCLUDED.attrs
        RETURNING symbol, xmax;
    """
    ).format(
        schema_name=sql.Identifier(Schema.SECURITY),
        table_name=sql.Identifier(AssetTbls.SYMBOLS),
        source=sql.Literal(source),
    )


# endregion


# region -------- -------- Timeseries Metadata Table Commands -------- --------
# Table to store metadata on each asset that has stored data. specifically time
# that their data starts and ends for each timeframe


def create_timeseries_metadata_view() -> sql.Composed:
    return sql.SQL(
        """
        CREATE MATERIALIZED VIEW IF NOT EXISTS {schema_name}.{view_name} AS
        WITH all_tables AS ({_available_tables} UNION ALL {_available_aggregates}),
        _metadata AS (
            SELECT
                dt.pkey,
                t.table_name,
                t.schema_name,
                t.timeframe,
                t.is_raw_data,
                t.trading_hours_type,
                dt.start_date,
                dt.end_date
            FROM all_tables t
            CROSS JOIN LATERAL get_timeseries_date_range(t.schema_name, t.table_name) AS dt
        )
        SELECT * FROM _metadata;
    """
    ).format(
        schema_name=sql.Identifier(Schema.SECURITY),
        view_name=sql.Identifier(AssetTbls._METADATA),
        _available_tables=_available_tables(),
        _available_aggregates=_available_aggregates(),
    )


def _available_tables() -> sql.Composable:
    # Sub-query to get all the table names
    # is_raw_data = True because it's a table. By nature is must be raw inserted data
    return sql.SQL(
        r"""
        SELECT 
            tablename AS table_name,
            schemaname AS schema_name,
            substring(tablename FROM '_(\d+)(?:_raw)?(?:_(ext|rth|eth))?$')::INT AS timeframe,
            TRUE AS is_raw_data,
            CASE 
                WHEN tablename ~ '_ext$' THEN 'ext'
                WHEN tablename ~ '_rth$' THEN 'rth'
                WHEN tablename ~ '_eth$' THEN 'eth'
                ELSE 'none'
            END AS trading_hours_type
        FROM pg_catalog.pg_tables 
        """
        "WHERE schemaname IN ({schemas}) "
        r""" AND tablename ~ '^[\D]+_\d+(_raw)?(_(ext|rth|eth))?$'
        AND tablename NOT LIKE 'pg\_%'
    """
    ).format(
        schemas=sql.SQL(", ").join(
            sql.Literal(s) for s in [Schema.TICK_DATA, Schema.MINUTE_DATA, Schema.AGGREGATE_DATA]
        )
    )


def _available_aggregates() -> sql.Composable:
    # Sub-query to get all the continuous aggregate names
    # is_raw_data = False because it's a continuous agg. By nature is must be derived.
    return sql.SQL(
        r"""
        SELECT 
            user_view_name AS table_name,
            user_view_schema AS schema_name,
            substring(user_view_name FROM '_(\d+)(?:_raw)?(?:_(ext|rth|eth))?$')::INT AS timeframe,
            FALSE AS is_raw_data,
            CASE 
                WHEN user_view_name ~ '_ext$' THEN 'ext'
                WHEN user_view_name ~ '_rth$' THEN 'rth'
                WHEN user_view_name ~ '_eth$' THEN 'eth'
                ELSE 'none'
            END AS trading_hours_type
        FROM _timescaledb_catalog.continuous_agg 
        """
        "WHERE user_view_schema IN ({schemas}) "
        r""" AND user_view_name ~ '^[\D]+_\d+(_raw)?(_(ext|rth|eth))?$'
        AND user_view_name NOT LIKE 'pg\_%'
        """
    ).format(
        schemas=sql.SQL(", ").join(
            sql.Literal(s) for s in [Schema.TICK_DATA, Schema.MINUTE_DATA, Schema.AGGREGATE_DATA]
        )
    )


def create_timeseries_metadata_subfunction() -> sql.Composed:
    "SQL Dynamic Function to get the Stored Date-Range of all Assets in a Timeseries Table"
    return sql.SQL(
        """
        CREATE OR REPLACE FUNCTION get_timeseries_date_range(_schema TEXT, _table_name TEXT)
        RETURNS TABLE(pkey INT, start_date TIMESTAMPTZ, end_date TIMESTAMPTZ) AS
        $$
        BEGIN
            RETURN QUERY EXECUTE format(
                'SELECT pkey, MIN(dt), MAX(dt) FROM %I.%I GROUP BY pkey',
                _schema, _table_name
            );
        END;
        $$ LANGUAGE plpgsql;
    """
    ).format()


def refresh_timeseries_metadata_view() -> sql.Composed:
    return sql.SQL("REFRESH MATERIALIZED VIEW {schema_name}.{view_name};").format(
        schema_name=sql.Identifier(Schema.SECURITY),
        view_name=sql.Identifier(AssetTbls._METADATA),
    )


def select_timeseries_metadata(
    filters: list[Filter] = [],
    rtn_args: list[str] = [
        "pkey",
        "table_name",
        "schema_name",
        "start_date",
        "end_date",
    ],
) -> sql.Composed:
    return select(Schema.SECURITY, AssetTbls._METADATA, rtn_args, filters)


# endregion



"Series Data Partial Class Functions"

from io import BytesIO
import logging
from typing import Iterable, Literal, Optional

from pandas import DataFrame, DatetimeIndex, Timedelta, Timestamp
import pandas

from psyscale.psql.timeseries import AGGREGATE_ARGS, TICK_ARGS, AggregateArgs, TickArgs
from psyscale.timeseries_partial import (
    TimeseriesPartialAsyncAbstract,
    TimeseriesPartialAbstract,
)
from .series_df import Series_DF
from .psql import (
    MetadataInfo,
    Operation as Op,
    Schema,
    AssetTbls,
    SeriesTbls,
    AssetTable,
)

log = logging.getLogger("psyscale_log")

# pylint: disable='protected-access','abstract-method'


class SeriesDataPartial(TimeseriesPartialAbstract):
    "Series Data Upsert and Fetch Functions"

    def get_series(
        self,
        symbol: int | str,
        timeframe: Timedelta,
        start: Optional[Timestamp] = None,
        end: Optional[Timestamp] = None,
        limit: Optional[int] = None,
        rth: bool = False,
        rtn_args: Optional[Iterable[AggregateArgs | TickArgs | str]] = None,
        *,
        mdata: Optional[MetadataInfo] = None,
    ) -> Optional[DataFrame]:
        """
        Fetch Series Data from the Database Aggregating the desired data as needed.

        -- PARAMS --
        - symbol : Int | Str
            - Symbol (str) or Primary Key (int) of the Symbol to Fetch (Case is ignored)
        - timeframe : pandas.Timedelta
            - Interval of the Data to Return. Doesn't not need to be a value stored in the
            database, merely one that can be derived from stored data.
            - Timedelta(0) will return Tick Data if it is stored for the given pkey
        - start : Optional pandas.Timestamp : Earliest Date of Data to Retrieve
        - end : Optional pandas.Timestamp : Latest Date of Data to Retrieve
        - limit : Optional Int : Maximum number of data points to return
        - rth : bool
            - When True, Return RTH Hours only
            - When False, Return All stored data, RTH/ETH/Closed/Breaks, etc
        - rtn_args : Optional list of arguments to return.
            - Default = {"dt", "open", "high", "low", "close", "volume", "price"}
            - Unknown args are ignored
            - Note: 'dt' will always be returned even if not included.

        - mdata : Optional MetadataInfo
            - Optional argument. This MetadataInfo is the result from calling inferred_metadata()
            for this pkey & timeframe. If not given, this function will retrieve it as needed.
        """
        if mdata is None:
            try:
                # Fetch the inferred metadata to get the table that will
                # be either aggregated, or copied from.
                mdata = self.inferred_metadata(symbol, timeframe, rth)
            except AttributeError:
                log.warning(
                    "Cannot Aggregate timeframe %s for symbol %s from the data in the database",
                    timeframe,
                    symbol,
                )
                return None

        if mdata is None:
            return None
        assert mdata.table

        # Configure return args as a set
        if rtn_args is None:
            _rtns = {"dt", "open", "high", "low", "close", "volume", "price"}
        else:
            # sql formatting functions remove excess/ unkown args
            _rtns = {*rtn_args}

        # Search Metadata for available data at TF or lower
        with self._cursor(dict_cursor=True) as cursor:
            if mdata.table.period != timeframe:
                log.info(
                    "Calculating Aggregate at Timeframe : %s, from %s Timeframe",
                    timeframe,
                    mdata.table.period,
                )
                cmd = self[Op.COPY, SeriesTbls.CALCULATE_AGGREGATE](mdata, timeframe, rth, start, end, limit, _rtns)
            else:
                # Works for both Aggregates and Raw Tick Data Retrieval
                cmd = self[Op.COPY, SeriesTbls.RAW_AGGREGATE](mdata, rth, start, end, limit, _rtns)

            buffer = BytesIO()
            with cursor.copy(cmd) as copy:
                buffer.writelines(copy)
            return _bytes_to_df(buffer)

    def upsert_series(
        self,
        pkey: int,
        metadata: MetadataInfo,
        data: DataFrame,
        exchange: Optional[str] = None,
        *,
        on_conflict: Literal["update", "error", "ignore"] = "ignore",
    ):
        """
        Insert or Upsert symbol data to the database.

        -- PARAMS --
        - pkey : int. Primary Key of the symbol to insert
        - metadata: MetadataInfo Object
            - Contains the schema_name & table_name to insert the data into, can be retrieved from
            calling 'symbol_series_metadata'
        - Data: Dataframe.
            - Should contain all series data needed to be inserted. Multiple names will be
            recognized for each given series parameter.
            - i.e. time/datetime/date/dt ... etc will all be recognized as the timestamp column.
        - exchange : str | None
            - Exchange that the Asset is traded on. This will be passed to pandas_market_calendars
            so the RTH/ETH session of each datapoint can be determined and stored as necessary.
            - None can be passed for 24/7 Exchanges such as Crypto. Note: Forex would require 24/5
        - on_conflict : Literal["update", "error", "ignore"] = "error"
            - Action to take when a UNIQUE conflict occurs. Erroring allows for faster insertion
            if it can be ensured that given data will be unique
        """
        table = metadata.table if metadata.table is not None else AssetTable.from_table_name(metadata.table_name)

        data = _configure_and_check_df(data, exchange, table)

        if on_conflict == "ignore":
            data = _filter_redundant_datapoints(data, metadata)
            if len(data) == 0:
                log.warning(
                    "Upsert_Symbol_Data() given only redundant data points and set to ignore.\n"
                    'No Data is being inserted. Set on_conflict="update" to edit existing data.'
                )
                return

        # Setup and copy data into database
        buffer_tbl_type = SeriesTbls.TICK_BUFFER if table.period == Timedelta(0) else SeriesTbls.RAW_AGG_BUFFER

        # Inject the data through a temporary table
        with self._cursor(raise_err=True) as cursor:
            # Create & Inject the Data into a Temporary Table
            cursor.execute(self[Op.CREATE, buffer_tbl_type](table))
            copy_cmd = self[Op.COPY, buffer_tbl_type](
                # Sends the COPY Cmd & the order of the Columns of the Dataframe
                [str(c) for c in data.columns]
            )
            with cursor.copy(copy_cmd) as copy:
                for row in data.itertuples(index=False, name=None):
                    # Writes each row as a Tuple that matches the Dataframe Column Order
                    copy.write_row(row)

            # Merge the Temp Table By inserting / upserting from the Temporary Table
            _op = Op.UPSERT if on_conflict == "update" else Op.INSERT
            cursor.execute(self[_op, buffer_tbl_type](metadata.schema_name, table, pkey))
            log.info("Symbol Data Upsert Status Message: %s", cursor.statusmessage)

        self._update_series_data_edit_record(metadata, data, table)

    def refresh_aggregate_metadata(self):
        """
        Refresh Continuous Aggregates & the Timeseries Metadata Table based on upserts made.
        Designed to be called after all known data insertions have been made.

        Edits made using 'upsert_symbol_data()' are tracked. This includes individual tables and the
        respective time-ranges edited. This method uses that stored information to update only what
        needs to be updated.

        CAVEAT: This only works so long as this is the same class instance that made the updates
        in the first place. If that instance is deleted before calling this function
        refresh_all_aggregates_and_metadata() must be invoked manually.
        """
        if not hasattr(self, "_altered_tables"):
            log.info("No Series Data has been inserted, Skipping Metadata Refresh.")
            return

        with self._cursor(auto_commit=True) as cursor:

            # Loop Through Schemas
            for schema, mdata_dict in self._altered_tables_mdata.items():
                log.info(" ---- ---- Refreshing Timeseries Schema : %s  ---- ---- ", schema)

                # Loop Through Edited Tables
                for table_name, mdata in mdata_dict.items():
                    log.info(
                        " --- Refreshing Aggregates Associated with Table : %s ---- ",
                        table_name,
                    )
                    assert mdata.table  # Ensuring mata.Table is defined by post_init
                    cont_aggs = self._table_config[Schema(schema)].get_tables_to_refresh(mdata.table)
                    # Add some buffer dates so entire time chucks are covered
                    # Times Chucks will not refresh unless they are completely included
                    mdata.start_date -= Timedelta("4W")
                    mdata.end_date += Timedelta("4W")

                    for table in cont_aggs:
                        if table.raw:
                            continue

                        log.info("Refreshing Continuous Aggregate : %s ", table.table_name)
                        cursor.execute(
                            self[Op.REFRESH, SeriesTbls.CONTINUOUS_AGG](schema, table, mdata.start_date, mdata.end_date)
                        )

            # Refresh the metadata View to Reflect Updates
            log.info("---- ---- Refreshing 'Security._Metadata' Materialized View ---- ----")
            cursor.execute(self[Op.REFRESH, AssetTbls._METADATA]())

        # Reset the mdata memory just in case
        del self._altered_tables
        del self._altered_tables_mdata

    def _update_series_data_edit_record(self, metadata: MetadataInfo, data: DataFrame, table: AssetTable):
        # Ensure records exist in this instance
        if not hasattr(self, "_altered_tables"):
            # pylint: disable=attribute-defined-outside-init
            self._altered_tables: dict[Schema | str, set[str]] = {}
            self._altered_tables_mdata: dict[Schema | str, dict[str, MetadataInfo]] = {}

        # Ensure the schema key exists in both dicts
        if metadata.schema_name not in self._altered_tables:
            self._altered_tables[metadata.schema_name] = set()
            self._altered_tables_mdata[metadata.schema_name] = {}

        # Update / Add the Necessary Metadata.
        if table in self._altered_tables[metadata.schema_name]:
            # Join the metadata keeping track of the full data-range edited
            mdata = self._altered_tables_mdata[metadata.schema_name][table.table_name]
            mdata.start_date = min(mdata.start_date, data.iloc[0]["dt"])
            mdata.end_date = max(mdata.end_date, data.iloc[-1]["dt"])
            self._altered_tables_mdata[metadata.schema_name][table.table_name] = mdata

        else:
            # Construct a new metadata instance to add to the records
            # Start / End Dates represent ranges that were updated/inserted
            mdata = MetadataInfo(
                -1,  # pkey irrelevant here. The table is what's tracked, not the symbol
                table.table_name,
                metadata.schema_name,
                data.iloc[0]["dt"],
                data.iloc[-1]["dt"],
                table,
            )
            self._altered_tables[metadata.schema_name].add(table.table_name)
            self._altered_tables_mdata[metadata.schema_name][table.table_name] = mdata


class AsyncSeriesDataPartial(TimeseriesPartialAsyncAbstract, SeriesDataPartial):
    "Async extension for Series Data Upsert and Fetch Functions"

    async def upsert_series_async(
        self,
        pkey: int,
        metadata: MetadataInfo,
        data: DataFrame,
        exchange: Optional[str] = None,
        *,
        on_conflict: Literal["update", "error", "ignore"] = "ignore",
    ):
        "See upsert_series() Docstring"
        table = metadata.table if metadata.table is not None else AssetTable.from_table_name(metadata.table_name)

        data = _configure_and_check_df(data, exchange, table)

        if on_conflict == "ignore":
            data = _filter_redundant_datapoints(data, metadata)
            if len(data) == 0:
                log.warning(
                    "Upsert_Symbol_Data() given only redundant data points and set to ignore.\n"
                    'No Data is being inserted. Set on_conflict="update" to edit existing data.'
                )
                return

        buffer_tbl_type = SeriesTbls.TICK_BUFFER if table.period == Timedelta(0) else SeriesTbls.RAW_AGG_BUFFER

        # Inject the data through a temporary table
        async with self._acursor(raise_err=True) as cursor:
            await cursor.execute(self[Op.CREATE, buffer_tbl_type](table))
            copy_cmd = self[Op.COPY, buffer_tbl_type](
                # Sends the COPY Cmd & the order of the Columns of the Dataframe
                [str(c) for c in data.columns]
            )
            async with cursor.copy(copy_cmd) as copy:
                for row in data.itertuples(index=False):
                    # You'd think that calling await millions of times
                    # would produce a lot of overhead.... somehow it doesn't
                    await copy.write_row(row)

            # Merge the Temp Table By inserting / upserting from the Temporary Table
            _op = Op.UPSERT if on_conflict == "update" else Op.INSERT
            await cursor.execute(self[_op, buffer_tbl_type](metadata.schema_name, table, pkey))
            log.info("Symbol Data Upsert Status Message: %s", cursor.statusmessage)

        self._update_series_data_edit_record(metadata, data, table)

    async def get_series_async(
        self,
        symbol: int | str,
        timeframe: Timedelta,
        start: Optional[Timestamp] = None,
        end: Optional[Timestamp] = None,
        limit: Optional[int] = None,
        rth: bool = False,
        rtn_args: Optional[Iterable[AggregateArgs | TickArgs | str]] = None,
        *,
        mdata: Optional[MetadataInfo] = None,
    ) -> Optional[DataFrame]:
        """
        Fetch Series Data from the Database Aggregating the desired data as needed.

        CAVEAT: While this does work and is Asynchronous, there is no way to decouple
        the Select Query & the COPY FROM operation. While the former would benefit
        from an await, awaiting the latter slows down the data retrieval a lot.
        Overall this function takes about 60% longer than the synchronous version.

        -- PARAMS --
        - symbol : Int | Str
            - Symbol (str) or Primary Key (int) of the Symbol to Fetch (Case is ignored)
        - timeframe : pandas.Timedelta
            - Interval of the Data to Return. Doesn't not need to be a value stored in the
            database, merely one that can be derived from stored data.
            - Timedelta(0) will return Tick Data if it is stored for the given pkey
        - start : Optional pandas.Timestamp : Earliest Date of Data to Retrieve
        - end : Optional pandas.Timestamp : Latest Date of Data to Retrieve
        - limit : Optional Int : Maximum number of data points to return
        - rth : bool
            - When True, Return RTH Hours only
            - When False, Return All stored data, RTH/ETH/Closed/Breaks, etc
        - rtn_args : Optional list of arguments to return.
            - Default = {"dt", "open", "high", "low", "close", "volume", "price"}
            - Unknown args are ignored
            - Note: 'dt' will always be returned even if not included.

        - mdata : Optional MetadataInfo
            - Optional argument. This MetadataInfo is the result from calling inferred_metadata()
            for this pkey & timeframe. If not given, this function will retrieve it as needed.
        """

        if mdata is None:
            try:
                # Fetch the inferred metadata to get the table that will
                # be either aggregated, or copied from.
                mdata = self.inferred_metadata(symbol, timeframe, rth)
            except AttributeError:
                log.warning(
                    "Cannot Aggregate timeframe %s for symbol %s from the data in the database",
                    timeframe,
                    symbol,
                )
                return None

        if mdata is None:
            return None
        assert mdata.table

        # Configure return args as a set
        if rtn_args is None:
            _rtns = {"dt", "open", "high", "low", "close", "volume", "price"}
        else:
            # sql formatting functions remove excess/ unkown args
            _rtns = {*rtn_args}

        async with self._acursor() as cursor:
            # Configure return args as a set
            if rtn_args is None:
                _rtns = {"dt", "open", "high", "low", "close", "volume", "price"}
            else:
                # sql formatting functions remove excess/ unkown args
                _rtns = {*rtn_args}

            if mdata.table.period != timeframe:
                log.info(
                    "Calculating Aggregate at Timeframe : %s, from %s Timeframe",
                    timeframe,
                    mdata.table.period,
                )
                cmd = self[Op.COPY, SeriesTbls.CALCULATE_AGGREGATE](mdata, timeframe, rth, start, end, limit, _rtns)
            else:
                # Works for both Aggregates and Raw Tick Data Retrieval
                cmd = self[Op.COPY, SeriesTbls.RAW_AGGREGATE](mdata, rth, start, end, limit, _rtns)

            buffer = BytesIO()
            async with cursor.copy(cmd) as copy:
                async for line in copy:
                    buffer.write(line)
            return _bytes_to_df(buffer)

    async def refresh_aggregate_metadata_async(self):
        """
        Refresh Continuous Aggregates & the Timeseries Metadata Table based on upserts made.
        Designed to be called after all known data insertions have been made.

        Edits made using 'upsert_symbol_data()' are tracked. This includes individual tables and the
        respective time-ranges edited. This method uses that stored information to update only what
        needs to be updated.

        CAVEAT: This only works so long as this is the same class instance that made the updates
        in the first place. If that instance is deleted before calling this function
        refresh_all_aggregates_and_metadata() must be invoked manually.
        """
        if not hasattr(self, "_altered_tables"):
            log.info("No Series Data has been inserted, Skipping Metadata Refresh.")
            return

        async with self._acursor(auto_commit=True) as cursor:

            # Loop Through Schemas
            for schema, mdata_dict in self._altered_tables_mdata.items():
                log.info(" ---- ---- Refreshing Timeseries Schema : %s  ---- ---- ", schema)

                # Loop Through Edited Tables
                for table_name, mdata in mdata_dict.items():
                    log.info(
                        " --- Refreshing Aggregates Associated with Table : %s ---- ",
                        table_name,
                    )
                    assert mdata.table  # Ensuring mata.Table is defined by post_init
                    cont_aggs = self._table_config[Schema(schema)].get_tables_to_refresh(mdata.table)
                    # Add some buffer dates so entire time chucks are covered
                    # Times Chucks will not refresh unless they are completely included
                    mdata.start_date -= Timedelta("4W")
                    mdata.end_date += Timedelta("4W")

                    for table in cont_aggs:
                        if table.raw:
                            continue

                        log.info("Refreshing Continuous Aggregate : %s ", table.table_name)
                        await cursor.execute(
                            self[Op.REFRESH, SeriesTbls.CONTINUOUS_AGG](schema, table, mdata.start_date, mdata.end_date)
                        )

            # Refresh the metadata View to Reflect Updates
            log.info("---- ---- Refreshing 'Security._Metadata' Materialized View ---- ----")
            await cursor.execute(self[Op.REFRESH, AssetTbls._METADATA]())

        # Reset the mdata memory just in case
        del self._altered_tables
        del self._altered_tables_mdata


def _configure_and_check_df(data: DataFrame, exchange: str | None, table: AssetTable) -> DataFrame:
    # region ---- Check that the data matches name and 'NOT NULL' expectations
    series_df = Series_DF(data.copy(), exchange)  # Rename cols & Populate 'rth'

    try:
        if table.period != Timedelta(0):
            # Aggregate Specific Expectations
            assert table.period == series_df.timedelta
            assert "close" in series_df.columns
            assert not series_df.df["close"].isna().any()
            extra_cols = set(series_df.df.columns).difference(AGGREGATE_ARGS)
        else:
            # Tick Specific Expectations
            assert "price" in series_df.columns
            assert not series_df.df["price"].isna().any()
            extra_cols = set(series_df.df.columns).difference(TICK_ARGS)

        if len(extra_cols) > 0:
            series_df.df.drop(columns=extra_cols, inplace=True)
            log.debug("Ignoring extra columns in dataframe: %s", extra_cols)

        # Regardless if Tick or aggregate, check the 'rth' to be NOT NULL when needed.
        if table.has_rth:
            assert "rth" in series_df.columns
            if (nans := series_df.df["rth"].isna()).any():
                drops = series_df.df[nans]
                log.warning(
                    "Edge-case mark session error. Dropping %s data-points: %s.",
                    len(drops),
                    drops,
                )
                series_df.df = series_df.df[~nans]
        elif "rth" in series_df.columns:
            extra_rows = series_df.df["rth"] != 0
            if len(extra_rows) > 0:
                log.warning(
                    "Given Extended hours data to an aggregate table that doesn't need it."
                    "Dropping extra data rows : %s",
                    extra_rows,
                )
                # drop all ext hours datapoints
                series_df.df = series_df.df[~extra_rows]
            series_df.df.drop(columns="rth")

        assert "dt" in series_df.columns
        assert not series_df.df["dt"].isna().any()
    except AssertionError as e:
        raise ValueError("Cannot insert symbol data into TimescaleDB. Dataframe is not formatted correctly") from e

    return series_df.df

    # endregion


def _filter_redundant_datapoints(data: DataFrame, metadata: MetadataInfo) -> DataFrame:
    # When inserting ensure only the range that needs to be added is.
    # Pretty common there's 1 extra data point at the start of a df.
    before_start = data["dt"] < metadata.start_date
    after_end = data["dt"] > metadata.end_date
    dt_filter = before_start | after_end

    if not dt_filter.all():
        extra_data = data[~dt_filter]
        log.debug(
            "Given %s extra data point(s), dropping the following :\n %s",
            len(extra_data),
            extra_data,
        )
        data = data[dt_filter]

    return data


def _bytes_to_df(buffer: BytesIO) -> DataFrame | None:
    buffer.seek(0)
    _rtn = pandas.read_csv(buffer)
    if len(_rtn) == 0:
        return None

    if _rtn["dt"].dtype == "int64":
        # Cursed? yes. Faster? Also yes.
        _rtn["dt"] = (_rtn["dt"] * 1e9).astype("datetime64[ns, UTC]")  # type: ignore
    else:
        _rtn["dt"] = DatetimeIndex(_rtn["dt"])
        # When returning tick data a timestamp is returned as a string.
        # its much slower, but retains the sub-second timestamp resolution
    return _rtn
